"""
Access sample NITF images from [1]_ for testing and demo purposes.

References:
    .. [1] https://gwg.nga.mil/ntb/baseline/software/testfile/Nitfv2_1/scen_2_1.html


NOTE:
    The data provided is no longer available, but a mirror exists on IPFS and
    the wayback machine.

    https://web.archive.org/web/20190501060607/http://www.gwg.nga.mil/ntb/baseline/software/testfile/Nitfv2_1/scen_2_1.html

    # IPFS folder containing all demo NITFs
    QmWApzbAX2W8cobWqKPjsM1kj82TvMGvyZ75PnbUeNHktW demo_nitf

    ipfs pin add --progress --name demo_nitf_data QmWApzbAX2W8cobWqKPjsM1kj82TvMGvyZ75PnbUeNHktW
    ipfs pin ls --type="recursive" --names
"""
import ubelt as ub


DEFAULT_KEY = 'i_3034c.ntf'  # use a small image as the default


_TEST_IMAGES = [
    {'key': 'ns3114a.nsf', 'sha512': '5605cd0b0187900c1b43130bf157c6ed', 'size_bytes': 680, 'enabled': False, 'CID': 'QmQKr5fd5cp1b1LYFgYRJKT6jdNHyecEN1XqRMKp4JJ4SM'},
    {'key': 'i_3034c.ntf', 'sha512': '5f42ab1034f20756bdf15043b58c828f', 'size_bytes': 933, 'CID': 'QmSRGeqnkmMGfXFDZS22yWt3TAbxun533kNCWjuw6rqqpC', 'alt_url': 'https://data.kitware.com/api/v1/file/62757dff4acac99f42ce826b/download'},
    {'key': 'ns3034d.nsf', 'sha512': 'aaf65232611bdc53934fa7f58790529a', 'size_bytes': 937, 'CID': 'QmStHdJpmTtmwP1cV589p8yMr9aGprukNfaKxGWHT2gXLk'},
    {'key': 'i_3034f.ntf', 'sha512': '2c2c3a918fe2805dc78a090164d671dc', 'size_bytes': 948, 'CID': 'QmUHHD92Fecr1wviDihGUkEomVoF7XHSPYgK2vZN26Et9i'},
    {'key': 'i_3051e.ntf', 'sha512': 'a67e9c4172310faaadf34a0e6ce01a72', 'size_bytes': 1436, 'enabled': False, 'CID': 'QmZLE6WudN5ZRCwpMDnpK43BnERds3qNPanzqgCpuX1BZC'},
    {'key': 'i_3052a.ntf', 'sha512': 'bb96983f58b3ec5891ef2f096d97b84e', 'size_bytes': 1520, 'enabled': False, 'CID': 'QmSkjd7DcqE34L8ThpAmpEPJNKa1ewSi2npQhrcjtreHmv'},
    {'key': 'ns3051v.nsf', 'sha512': 'ca438abdb6e67776a99bd14cc4154fa6', 'size_bytes': 1592, 'enabled': False, 'CID': 'QmRDMx5yDx8eBWFggj8556ke5iCtuvymLRfhBRonS2usba'},
    {'key': 'i_3063f.ntf', 'sha512': '68bb22e59739f31c1f5cc66cc5ac415b', 'size_bytes': 1596, 'enabled': False, 'CID': 'QmUFrYdz5xHrvMbcEWCpMurrqbXago7RK5U76EfnUrdniV'},
    {'key': 'ns3063h.nsf', 'sha512': '7b85023c46414b4b36c2a2795e59e2bd', 'size_bytes': 1606, 'enabled': False, 'CID': 'QmcFNheBVnjUVakDi5So5UtpurAaPXxYwdEj6pxqRCX57e'},
    {'key': 'i_3060a.ntf', 'sha512': '9e42e2d8e9fe6c07380501fcf5339c3e', 'size_bytes': 1624, 'enabled': False, 'CID': 'QmWCRDnbEkX7Nqju5vfUjetFTDek5GLHLjg2UwvHMsuffy'},
    {'key': 'i_3068a.ntf', 'sha512': 'aa4913146f4512b41ba2d6717a375a94', 'size_bytes': 1658, 'enabled': False, 'CID': 'QmbxM3NB8iFJs5gQk2QJi4mFHbfFBcexpJEN6WsKq7Rq19'},
    {'key': 'ns3061a.nsf', 'sha512': '86013922c8caf816943296a8b3dcbbd5', 'size_bytes': 1668, 'enabled': False, 'CID': 'QmUfNxVJttBDsJcjeQWpk8pY1FgxHkGbepy3WC4zsVTAjs'},
    {'key': 'ns3059a.nsf', 'sha512': 'c84eed6be08e60921920b4bfe621e0b6', 'size_bytes': 1766, 'enabled': False, 'CID': 'QmdjWZHsjBKa5YKTcNBaEuxZb1tXzrVy3JSnXfHCGjomBA'},
    {'key': 'i_3114e.ntf', 'sha512': '709933f92fb86a0022fed6f4edd32626', 'size_bytes': 1776, 'enabled': False, 'CID': 'QmcAakF7RN533TiU8qnhGSUoKZuScCDXZVLeJmdcq7C9PG'},
    {'key': 'ns3073a.nsf', 'sha512': 'a70636dcc455307f4cea235fb05d7ddd', 'size_bytes': 1854, 'enabled': False, 'CID': 'QmXUjPur9Cbe1sNgKkPP3sBfqkWfymKuiehNqHu8WuFU3A'},
    {'key': 'ns3101b.nsf', 'sha512': '7535d935a56ac1d1a7ac85e3d6c03983', 'size_bytes': 2144, 'enabled': False, 'CID': 'QmTjPsxGvQzdvH9T61ewEuoJKkczRL8U4CSTnqJuZhH6DL'},
    {'key': 'i_3025b.ntf', 'sha512': '582bb41bd308535ef81cc822ad68d90f', 'size_bytes': 2199, 'CID': 'QmfKK7NPfggEJAmVfv8VfFrzL6P8rpxTBUtX33YhHqmRF5'},
    {'key': 'i_3076a.ntf', 'sha512': '97200f22508e02e130799fedc64bf2a2', 'size_bytes': 2246, 'enabled': False, 'CID': 'QmUKSJRYKKDhsvZwdFgPnHLwzbEB22TFwwsoGxwCtB1Wne'},
    {'key': 'i_3018a.ntf', 'sha512': '6e8539c992f289ed7292401ce4a394a1', 'size_bytes': 3564, 'CID': 'QmfVBure4Jbmo7ESt5aotYbr5MnesWJxMFfRVLRoHqAfDd'},
    {'key': 'ns3050a.nsf', 'sha512': '1a87ae99cd1952360159a43e9f0fd269', 'size_bytes': 4071, 'CID': 'QmVQgiFEV1w1GGw9AM3abNqU3fhPFbLTrgSLsbGXPzhWXG'},
    {'key': 'ns3022b.nsf', 'sha512': '876d0e4435132cb667ef7f7b662e4ac2', 'size_bytes': 4502, 'CID': 'QmT3TpNozoLqS96dA5WCiWyZU1j3CwiAtkPmnU4WFNNe5B'},
    {'key': 'i_3015a.ntf', 'sha512': '6f7e8c8e5c93f3bfe4795dafb194cd13', 'size_bytes': 6074, 'CID': 'QmPaDjuweukFaDiM44rxkHaSMwcvLu5gM6U8EY2XU6XzAF'},
    {'key': 'ns3038a.nsf', 'sha512': '77e3270feb072d8173547cde48b8128c', 'size_bytes': 7018, 'CID': 'QmYG9PdWmJLRxZj6Qesv4uAj5gjeq3WZMofaDCQZKRsbaa'},
    {'key': 'ns3417c.nsf', 'sha512': 'd28b53eee6afb3eee010447dbf3b952a', 'size_bytes': 8836, 'enabled': False, 'CID': 'QmQDsLkqiZ2yKbF7qdShEZupiAuuzYsL7UJQoWiakJ6zbL'},
    {'key': 'ns3010a.nsf', 'sha512': 'dddede2a6bc30d9dad64ee062668b14a', 'size_bytes': 10711, 'CID': 'QmS8zETJ9fpJqAjSSiTMbTMEDcZJPjs5LG3EMhdGwYURef'},
    {'key': 'ns3017a.nsf', 'sha512': 'f074fe40a4ddfee2627b2561f6c8f360', 'size_bytes': 11052, 'CID': 'QmNTmkLffWuu4dKXz7WhAe99ErtvFMpiaogLQaRVDy3GSt'},
    {'key': 'i_3008a.ntf', 'sha512': '3a6a693f2ad48de56494017d0cb7cd5e', 'size_bytes': 21212, 'CID': 'QmVEfXgbQrHYVYbjiek5TEvxmvyWuLSLas5vnCmxBCJ5CG'},
    {'key': 'ns3114i.nsf', 'sha512': '3381b75a976da70022eb1564e07b5a73', 'size_bytes': 26664, 'enabled': False, 'CID': 'QmS3rjo6yuhGVVMtXpjkoR8HhbzXv9B7nJ3CbE7uZedSrj'},
    {'key': 'i_3201c.ntf', 'sha512': '55789ec5e5cf772c7b186bd6a0b57d6b', 'size_bytes': 48497, 'CID': 'QmNhrFM4SYsqUk333MCdTgzzwN5j1nvmCAu2Kjmq3ohEDm'},
    {'key': 'ns3033b.nsf', 'sha512': '85633cb72f4c10fcf8c98f6caffd5b9b', 'size_bytes': 53061, 'CID': 'QmXjnNXcAEwje3VFu6VYgVeeb9PacD6NRSRamV6Pr8z33s'},
    {'key': 'i_3041a.ntf', 'sha512': 'bdb9215995a602231003caf7aca0d3b7', 'size_bytes': 64682, 'CID': 'QmdNAsELfULbtXHTZhmfQ1svZVY24thttPpn4GNCZ9M2PE'},
    {'key': 'i_3113g.ntf', 'sha512': '143d866fb1bc4ee5e16b458f5e8c1e24', 'size_bytes': 70765, 'enabled': False, 'CID': 'QmRSrjcm75KiCw5zjJFvWGJXSKaLLvRNpj7y8Bqd67uBnk'},
    {'key': 'ns3301j.nsf', 'sha512': '813992b3600624b9748847180a007889', 'size_bytes': 95605, 'CID': 'QmW6Dpgrr2x94DGse6rRVuyAajMPnN7r6U9UgmEzeEQ8Uz'},
    {'key': 'ns3005b.nsf', 'sha512': '49702adf3d64648b8a29b7ef7ca7a132', 'size_bytes': 129920, 'CID': 'QmViVzFPjWe7KEF1eFDniBDggRrd2V7vAomnghyvtWYjbm'},
    {'key': 'i_3301h.ntf', 'sha512': 'faf89333207b840e11ac06b44b0ff09c', 'size_bytes': 140837, 'CID': 'QmSjNfgd2iue4ZoPbhzoTcp7isVK1aVBcDbPqYFsdRwTxg'},
    {'key': 'ns3201a.nsf', 'sha512': 'ab5cc412ddfba05463f5d84593c726da', 'size_bytes': 170590, 'CID': 'QmTibUeWxAW2YpC5WL6egku4aq95VfXTGC6CiJfcX3KLaJ'},
    {'key': 'ns3302a.nsf', 'sha512': 'c3445af2d5f529e42be7d66bde52311c', 'size_bytes': 197477, 'CID': 'QmVysZ9HexrwZ131U8iu4HVXgBtVGXazcJupA3yU7avp7n'},
    {'key': 'ns3310a.nsf', 'sha512': '62cec9952b7d98de3f550e410099ab93', 'size_bytes': 197477, 'CID': 'QmfMzFPqdeYvqpk33qRDsuhkzQbqydqXgQkvCMynMAtAnJ'},
    {'key': 'ns3301e.nsf', 'sha512': '2659e9f504b2427ebe064fa7bbfceeea', 'size_bytes': 197504, 'CID': 'QmVMwtyKUzzeUiHvNPWWhZV91rJbngy9F22wqkdFBBDUpW'},
    {'key': 'ns5600a.nsf', 'sha512': 'f7e27013e66568e02b0c4039120ff155', 'size_bytes': 219974, 'CID': 'Qmf6HT32uBxgPnYanYnaBrDyQdob9FuM5dv2mDSDejBxvK'},
    {'key': 'i_3128b.ntf', 'sha512': 'e4fa986099dcfdaaa74e66f68f48abbc', 'size_bytes': 248762, 'CID': 'QmQrnGbJ8XbUWipiw582WHbvpPqF2gLZsHj2tvRFgudC7H'},
    {'key': 'i_3004g.ntf', 'sha512': '5285370808d15ffcbc57a53105ff7a1b', 'size_bytes': 263047, 'has_crs': True, 'has_rpc': False, 'CID': 'QmSZkJKidBsWAUyPRFRbHFYxy1d4DBArhXD8utgavNASDp',
     'alt_url': 'https://data.kitware.com/api/v1/file/627579e94acac99f42ce69b1/download'},
    {'key': 'ns3004f.nsf', 'sha512': '0bfbd6d378dfd0f0e8cad703498fa6c8', 'size_bytes': 263047, 'has_crs': True, 'has_rpc': False, 'CID': 'QmXNq8HGVk5cMRTPSc9br685tpJqhBvky8Wmd6KGBKGA6t'},
    {'key': 'ns3090i.nsf', 'sha512': '9be0244363bbe63c71b55217a90c346b', 'size_bytes': 264083, 'enabled': False, 'CID': 'QmPfNJhP4Ytgx6D566NTt4CYZG2FJDamibNZaZYSLjDVG1'},
    {'key': 'i_3090m.ntf', 'sha512': '30017da0f1c9c41130e79c18f99aba97', 'size_bytes': 264083, 'CID': 'QmXhiB3uFeNHT91PFSbSNEV3zbDrkGKr9FTGA3GeVrMbfU'},
    {'key': 'i_3090u.ntf', 'sha512': 'a3f57adee4e5e25f03131891e6948da4', 'size_bytes': 264091, 'CID': 'QmYcA5UuwpMu6C7T92tLMMpmkMR69fzv3BdNt7y5KfDnTk'},
    {'key': 'ns3090q.nsf', 'sha512': '3c4f69ed8298f40e9a4845ae8321b056', 'size_bytes': 264091, 'CID': 'QmZBz5vKwygBtdfmRDG6xrbG2RtU8HWf5L7aAP5h4AF5gV'},
    {'key': 'ns3361c.nsf', 'sha512': '67123e051a8d02909e6b53b703330db9', 'size_bytes': 264592, 'has_crs': True, 'has_rpc': False, 'CID': 'QmSQBWmZm6WNzzuSySEYvSEiqKt3SpPTuuevVSjQh6g3DT'},
    {'key': 'ns3321a.nsf', 'sha512': '5a82d19b8a903537bee14e4d3a7cdc55', 'size_bytes': 281130, 'CID': 'Qmf9LVfsMTyFfvLpcXJAqCkZEoGG5KFA4QX8BoaXnz8JjX'},
    {'key': 'ns3118b.nsf', 'sha512': 'c95f99cf7bdd0ae2f802a282b772e339', 'size_bytes': 362407, 'CID': 'QmRmjCMyhHYBN5WfXJteTnAxYoJSQ9iK3KWbjDTe6DoTav'},
    {'key': 'i_5012c.ntf', 'sha512': 'de025f0a3da3b4f9e7279a45b7cd02e5', 'size_bytes': 594601, 'CID': 'QmbgWEVKE78tspn7ZYnmtPs2QYZFpirtEQxGYNQJkVghYw'},
    {'key': 'ns3304a.nsf', 'sha512': '5b0418c2f2cae7038eebdf3764514129', 'size_bytes': 701687, 'enabled': False, 'CID': 'QmUq7tL1EH2KyByNCFvRQnrCUkds4ZjpoPi78YKc1ZA94o'},
    {'key': 'i_3309a.ntf', 'sha512': '099d017dbfee8f703c4e6d76b9810a0e', 'size_bytes': 722432, 'CID': 'QmTgPnfTaHiz8zma8jDQ9n72yVeWLqApihmgQWrA4Q5wpk'},
    {'key': 'i_3001a.ntf', 'sha512': 'e5cdb23c612cbe28f0b994b69285aa49', 'size_bytes': 1049479, 'has_crs': True, 'has_rpc': False, 'CID': 'QmfNRAZdn3VtEHQSzXRYdsFCZozMSVUuHv85yjyY6EPmka'},
    {'key': 'ns3119b.nsf', 'sha512': '14b7ee574116538f88b0a7aa7758c88b', 'size_bytes': 1051108, 'enabled': False, 'CID': 'QmZzY3Vn93xKv5KEh7cSRLUEjFx29k8pxwD49e7fA4zfUx'},
    {'key': 'i_3430a.ntf', 'sha512': 'c52c72adf654c5e02fb5bb19174c5e99', 'size_bytes': 1573707, 'CID': 'QmU58g5BUDUc6k5UXX8CBgjUcR6yhvxLwicjdqMWkbADXj'},
    {'key': 'i_3301k.ntf', 'sha512': 'daf36eec7d4eb4be006c6f098d0efea8', 'size_bytes': 1770388, 'CID': 'QmZP22WEv6X3Mr5XeyZDP1XeX5u7w7RrHhD3U3de5kbvo9'},
    {'key': 'ns3301b.nsf', 'sha512': '9a51ed8ec667b2618d8d1f357a84b1e5', 'size_bytes': 1770388, 'CID': 'QmQxKrj1SppQcXHhAJBv4GccQxjdXfSf8psWH5qAjXnqLc'},
    {'key': 'i_3301c.ntf', 'sha512': '07de02322f46f9ba90318fb97b8ca759', 'size_bytes': 1770460, 'CID': 'QmYYwyHfQnx4jdpsXuwQU1jTtHVWeaLsCq7ftFP6PkLP7H'},
    {'key': 'i_3405a.ntf', 'sha512': 'fde7c7c42b42ce3bf6f6193f9108edf1', 'size_bytes': 2097995, 'CID': 'QmTYF1Hn9wAUoo8aTbEnnr9ty2NCTAnC82Sys7My9D38mi'},
    {'key': 'i_3450c.ntf', 'sha512': 'bea077c7d21d85e69f87fb4e565388d6', 'size_bytes': 2097995, 'CID': 'QmRiwn3nDxTmsjwiKgbyztMT7vwVqaKo2NP2N7Y42wy5MY'},
    {'key': 'ns3450e.nsf', 'sha512': 'ffa6e6923a2573d0aefb52908f078762', 'size_bytes': 2097995, 'CID': 'QmQSNZY3kGTxSCGw3gJftrqUHjdpDnnq9hJL5BFq4aRxbu'},
    {'key': 'ns3437a.nsf', 'sha512': '85a1bdcda593373326e5a0a498a79f95', 'size_bytes': 2656587, 'enabled': False, 'CID': 'QmQhKDL4S7mr6WkHPx9g2YgpUodUNq5nmYkSEzfKcispYa'},
    {'key': 'i_3301a.ntf', 'sha512': 'f669994ddaab9f08e1064f24dc0e1580', 'size_bytes': 3146597, 'enabled': False, 'CID': 'QmWqspPteMi38xHoohYLAhSqbcdNepWNNvS4f711uzicp8'},
    {'key': 'i_3117ax.ntf', 'sha512': 'c9ab95cc2cd4711677a0cce78122b703', 'size_bytes': 3489726, 'CID': 'QmYw4wppfG4uARMcaNrfrCCwrnTQmBb7cpW7VNZFRM7sXF'},
    {'key': 'i_3303a.ntf', 'sha512': 'e13a7aa57775b71d10e5a420e7a13214', 'size_bytes': 4195147, 'enabled': False, 'CID': 'QmVmCk77Dw4sTHvHKRBDxsto4RDQ1FQrf2roS8ntrALuCv'},
    {'key': 'i_3311a.ntf', 'sha512': '74e58d3b921555544678ae3488cb6a35', 'size_bytes': 4679168, 'CID': 'QmeQ3y7pAVgP7KxNwwaUuJyZQjvryL4yTisGneAoPS1eyJ'},
    {'key': 'ns3229b.nsf', 'sha512': '1004f108fd4b2841d3f7362ae4077e28', 'size_bytes': 5659571, 'CID': 'QmTJHhqvnMNnTMhDkq3rbRvQ8Bw6by5UUYvH8FrBUTxCqQ'},
    {'key': 'ns3228b.nsf', 'sha512': '2e9445fa3876e2e09aaa362f25f3018d', 'size_bytes': 6292578, 'CID': 'QmZuqBAUYEtYiRZ8hHoDJ3iYRf86UzcJHfSh5f6HLjmveU'},
    {'key': 'i_3228c.ntf', 'sha512': '2b059b564911c0f7c93b3ab3e332e480', 'size_bytes': 6292578, 'CID': 'QmeKe7twmNEq5iFTF6RZegVtuLBnm2GaVSnfLdbtLRkanh'},
    {'key': 'i_3228e.ntf', 'sha512': '3d9814143e2281241923904c4132859c', 'size_bytes': 6292578, 'CID': 'QmYwNVAaC9bZNCN6T5Qjojjsub2C2sVWy7y5RD3ah7VtA2'},
    {'key': 'ns3228d.nsf', 'sha512': 'a9b4ebab56101935eccc0b27b1060810', 'size_bytes': 6292578, 'CID': 'QmRrdi31As3dguMeGUZC1CroMmuj4cD3AiBmZXyWJYX6wg'},
]


_FNAME_TO_INFO = {row['key']: row for row in _TEST_IMAGES if row.get('enabled', True)}


_FNAME_TO_DESC = {
    'i_3001a.ntf': 'Can the system handle an uncompressed 1024x1024 8-bit mono image and file contains GEO data? (AIRFIELD)',
    'i_3004g.ntf': 'Checks a system to see how it applies GEO data around 00, 180.',
    'i_3008a.ntf': 'Checks a JPEG-compressed, 256x256 8-bit mono image, Q4, COMRAT 00.4 with general purpose tables embedded. File also contains image comments. (TANK)',
    'i_3015a.ntf': 'Can the system handle a JPEG-compressed 256x256 8-bit mono image with comment in the JPEG stream before frame marker? (TANK)',
    'i_3018a.ntf': 'Checks a JPEG-compressed 231x191 8-bit mono image with a corrupted restart marker occurring too early. (BLIMP)',
    'i_3025b.ntf': 'Checks to see if a viewer can read a JPEG stream with fill bytes (FF) in the JPEG stream before FFD8. (LINCOLN)',
    'i_3034c.ntf': 'Checks a 1-bit RGB/LUT with an arrow, the value of 1 mapped to green and the background value of 0 mapped to red, and no mask table.',
    'i_3034f.ntf': 'Checks a 1-bit RGB/LUT (green arrow) with a mask table (pad pixels having value of 0x00) and a transparent pixel value of 1 being mapped to green by the LUT.',
    'i_3041a.ntf': 'Checks a bi-level compressed at 2DS 512x512 FAX image. (SHIP)',
    'i_3051e.ntf': 'Checks to see if a system can render CGM Text in the proper location.',
    'i_3052a.ntf': 'Checks to see if the system renders a basic Circle.',
    'i_3060a.ntf': 'Checks for rendering CGM polylines (types 1 through 5.)',
    'i_3063f.ntf': 'Checks for rendering CGM polygons with hatch style 5.',
    'i_3068a.ntf': 'Checks for rendering CGM rectangles with starting point in Lower Right of rectangle.',
    'i_3076a.ntf': 'Checks for rendering various CGM elliptical arc cords.',
    'i_3090m.ntf': 'CIRARCC5 checks for proper interpretation of upper left VDC and drawing of center-closed CGM circular arcs across different quadrants.',
    'i_3090u.ntf': 'CIRARCCD checks for proper interpretation of upper right VDC and drawing of center-closed CGM circular arcs across different quadrants.',
    'i_3113g.ntf': 'Can system display a Low Bite Rate (LBR) file with an uncompressed image overlay?',
    'i_3114e.ntf': 'Checks to see if the system recognizes all UT1 values 0xA0 to 0xFF.',
    'i_3117ax.ntf': 'Can the system render an NSIF file having the maximum total bytes in 32 text segments each of 99,998 bytes with an image segment? (Text shows 1 of 32 identical text segments.)',
    'i_3128b.ntf': 'This file contains PIAE TREs version 2.0 to include three PEA TREs. If the system supports PIAE TREs, can they find each TRE to include all 3 PEA TREs?',
    'i_3201c.ntf': 'Checks a systems ability to handle a single block IMODE R image, 126x126',
    'i_3228c.ntf': 'MS IMODE P RGB, multi-blocked image, not all bands displayed.',
    'i_3228e.ntf': 'MS IMODE R RGB, multi-blocked image, not all bands displayed.',
    'i_3301a.ntf': 'Checks an uncompressed 1024x1024 24-bit multi-blocked (IMode-S) color image. (HELO)',
    'i_3301c.ntf': 'Checks an IMODE S image with a data mask subheader, the subheader with padded pixels, having a pad pixel value of 0x00 displaying as transparent, 3x3 blocks.',
    'i_3301h.ntf': 'Can the system display a multi block 6x6 IMODE R image and 216x216?',
    'i_3301k.ntf': 'Checks an IMODE R image with a data mask subheader, with padded pixels, a pad pixel value of 0x00 displaying as transparent, and 3x3 blocks.',
    'i_3303a.ntf': 'Can the system display an uncompressed 2048x2048 8-bit multi-blocked mono image? (CAMELS)',
    'i_3309a.ntf': 'Can the system display a JPEG-compressed 2048x2048 8-bit multi-blocked (256x256) mono image w/QFAC=3, RSTI=16, and IMODE=B? (CAMELS)',
    'i_3311a.ntf': 'Can the system display a JPEG 2048x2048 24-bit PI block color w/QFAC=3,RSTI=32,IMODE=P, blocked (512x512)? (JET)',
    'i_3405a.ntf': 'Can the system handle a multi-blocked 1024x1024 image with 11/16 (ABPP=11, NBPP=16)? (AIRSTRIP)',
    'i_3430a.ntf': 'Can the system handle an NSIF file with an uncompressed image with 12-bit back to back data, ABPP = 12, and NBPP = 12?',
    'i_3450c.ntf': 'Can the system read a 32-bit real image?',
    'i_5012c.ntf': 'Can the system handle an NSIF file with 100 images, 100 symbols and 32 text elements, images 1, 25, 50, 75 and 100 attached to "000", symbol 12 and text 29 attached to image 25, symbol 32 and text 30 attached to image 50, symbol 86 and text 31 attached to image 75, symbol 90 and text 32 attached to image 100, and all other segments attached to image 1?',
    'ns3004f.nsf': 'Checks a system to see how it applies GEO data around 00, 000.',
    'ns3005b.nsf': 'Checks a JPEG-compressed 1024x1024 8-bit mono image compressed with visible 8-bit tables and COMRAT 01.1. (AIRFIELD)',
    'ns3010a.nsf': 'Can the system handle a JPEG-compressed 231x191 8-bit mono image that is non-divide by 8, and file also contains image comments? (BLIMP)',
    'ns3017a.nsf': 'Checks a JPEG-compressed 231x191 8-bit mono image with a corrupted restart marker occurring too late. (BLIMP)',
    'ns3022b.nsf': 'Checks a JPEG-compressed 181 x 73 8-bit mono image with split Huffman tables 1 DC 1 AC having separate marker for each. (JET)',
    'ns3033b.nsf': 'Checks a JPEG-compressed 512x512 8-bit mono image with APP7 marker in JPEG stream. (LENNA)',
    'ns3034d.nsf': 'Checks a 1-bit mono with mask table having (0x00) black as transparent with white arrow.',
    'ns3038a.nsf': 'Checks all run lengths on a bi-level compressed at 1D and 1024x1024 FAX imagery. (SEMAPHORE)',
    'ns3050a.nsf': 'Checks all run lengths on a bi-level compressed at 2DH and 1024x1024 FAX imagery. (SEMAPHORE)',
    'ns3051v.nsf': 'Checks to see if the system can render CGM polygon sets properly and two polygons that do not intersect.',
    'ns3059a.nsf': 'Checks for rendering CGM ellipses with edge width of 50.',
    'ns3061a.nsf': 'Checks an IMODE S image with a data mask subheader, the subheader with padded pixels, having a color value of 0x00, 0x00, 0x00 displaying as transparent, and 3x3 blocks.',
    'ns3063h.nsf': 'Checks for rendering CGM polygons with hatch style 1 with auxiliary color.',
    'ns3073a.nsf': 'Checks for rendering various CGM circular arcs.',
    'ns3090i.nsf': 'CIRARCC1 checks for proper interpretation of lower left VDC and drawing of center-closed CGM circular arcs across different quadrants.',
    'ns3090q.nsf': 'CIRARCC9 checks for proper interpretation of lower right VDC and drawing of center-closed CGM circular arcs across different quadrants.',
    'ns3101b.nsf': 'Checks to see what CGM fonts are supported by the system. The display image is shown with limited font support.',
    'ns3114a.nsf': 'Can the render an NSIF file with a single (STA) text segment with only one byte of data?',
    'ns3114i.nsf': 'Can the system render a U8S character set (this text segment is in an HTML format)? (To verify data, ensure your web browser is set to properly display Unicode UT8-F.)',
    'ns3118b.nsf': 'Can the system render an embedded MTF file is the second text segment. Text shows MTF text segment.Can the system render an embedded MTF file that is the second text segment? (Text shows MTF text segment.)',
    'ns3119b.nsf': 'Can the system render the maximum CGM total bytes for a clevel 3 file (total bytes 1,048,576 in 8 CGM segments)?',
    'ns3201a.nsf': 'Checks a systems ability to handle an RGB/LUT. (LUT has 128 entries.)',
    'ns3228b.nsf': 'MS IMODE S RGB, multi-blocked image, not all bands displayed.',
    'ns3228d.nsf': 'MS IMODE B RGB, multi-blocked image, not all bands displayed.',
    'ns3229b.nsf': 'Nine band MS image, PVTYPE=SI, ABPP=16 in NBPP=16, IMODE B. Band 1, 2 & 3 have been enhanced for viewing, image is naturally dark.',
    'ns3301b.nsf': 'Checks an IMODE B image with a data mask subheader, the subheader with padded pixels, having a pad pixel value of 0x00 displaying as transparent, 3x3 blocks.',
    'ns3301e.nsf': 'Checks an IMODE P image with a data mask subheader, the subheader with padded pixels, having a pad pixel value of 0x7F displaying as determined by the ELT, 4x4 blocks.',
    'ns3301j.nsf': 'Can the system display a mono JPEG image with mask blocks?',
    'ns3302a.nsf': 'Can the system display an uncompressed 256x256 24-bit multi-blocked (IMode-B) image? (TRACKER)',
    'ns3304a.nsf': 'Can the system display a JPEG-compressed 2048x2048 8-bit multi-blocked (512x512) mono image w/QFAC=3, RSTI=32, and IMODE=B? (CAMELS)',
    'ns3310a.nsf': 'Can the system display an uncompressed, 244x244 24-bit IMODE P multi-blocked (128x128) color image? (BIRDS)',
    'ns3321a.nsf': 'Can the system handle an NSIF file containing a streaming file header (in which the image size was unknown at the time of production) and the main header has replacement data?',
    'ns3361c.nsf': 'How does the system handle multi-images with GEO data?',
    'ns3417c.nsf': 'Can the system handle a 98x208 mono image with custom 12-bit JPEG SAR tables and COMRAT 03.5?',
    'ns3437a.nsf': 'Can the system handle a 12-bit JPEG C5 (Lossless) ES implementation multi-blocked 1024x2048 image with APP6 in each displayable block?',
    'ns3450e.nsf': 'Can the system read a 64-bit real image?',
    'ns5600a.nsf': 'Can the system handle a MS, 31 Band image, 42 by 42 pixels, 32bpp Float, and IREPBANDS all blank?',
}


def grab_nitf_fpath(key=None):
    """
    Args:
        key (str | None): the name the nitf to grab.
            Use ``grab_nitf_fpath.keys()`` to list available keys.
            If None, ``DEFAULT_KEY`` is used.

    Example:
        >>> # xdoctest: +SKIP
        >>> # xdoctest: +REQUIRES(--network)
        >>> from kwgis.demo.nitf_demodata import *  # NOQA
        >>> fpath = grab_nitf_fpath()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> data = kwimage.imread(fpath)
        >>> kwplot.imshow(data)
        >>> kwplot.show_if_requested()

    Ignore:
        for key in grab_nitf_fpath.keys():
            fpath = grab_nitf_fpath(key)
    """
    # base = 'https://gwg.nga.mil/ntb/baseline/software/testfile/Nitfv2_1/'
    if key is None:
        key = DEFAULT_KEY

    if key in _FNAME_TO_INFO:
        info = _FNAME_TO_INFO[key]
    else:
        raise KeyError(key)

    if 'alt_url' in info:
        try:
            fname = info['key']
            sha512 = info['sha512']
            fpath = ub.grabdata(info['alt_url'], appname='kwgis/demodata/nitf',
                                fname=fname, hash_prefix=sha512)
            return fpath
        except Exception:
            pass

    if 'CID' in info:
        # Use IPFS instead
        # https://ipfs.github.io/public-gateway-checker/
        ipfs_gateways = [
            'https://ipfs.io/ipfs',
            'https://dweb.link/ipfs',
            'https://gateway.pinata.cloud/ipfs',
        ]
        url_fname = info['CID']
        fname = info['key']
        sha512 = info['sha512']
        import urllib
        fpath = None
        # try different gateways if the first one times out
        for try_idx, gateway in enumerate(ipfs_gateways):
            url = gateway + '/' + url_fname
            try:
                fpath = ub.grabdata(url, appname='kwgis/demodata/nitf',
                                    fname=fname, hash_prefix=sha512)
            except urllib.error.HTTPError as ex:
                print('caught error ex = {!r}'.format(ex))
                if try_idx == len(ipfs_gateways) - 1:
                    # ipfs_exe = ub.find_exe('ipfs')
                    # print(f'ipfs_exe={ipfs_exe}')
                    # if ipfs_exe:
                    #     fpath = ub.Path.appdir('kwgis/demodata/nitf') / fname
                    #     print(f'fpath={fpath}')
                    #     ub.cmd([ipfs_exe, 'get', '-o', str(fpath), str(url_fname)], verbose=3)
                    # else:
                    raise
                else:
                    print('Try again...')
        if fpath is None:
            raise AssertionError('should not happen')
    else:
        url_fname = info['key']
        raise Exception('Requires update to IPFS support. Old URLs are dead')
    return fpath


grab_nitf_fpath.keys = _FNAME_TO_INFO.keys


def _dev_build_description_table():
    """
    Developer function used to help populate data in this file.
    Unused at during runtime.

    Notes:
        No longer works, the URL is dead!

    Requirements:
        !pip install bs4
    """
    import bs4
    import requests
    resp = requests.get('https://gwg.nga.mil/ntb/baseline/software/testfile/Nitfv2_1/scen_2_1.html', verify=False)
    soup = bs4.BeautifulSoup(resp.text, 'html.parser')
    tables = soup.findAll('table')

    name_to_fname = {row['key'].split('.')[0]: row['key'] for row in _TEST_IMAGES}
    names_noext = list(name_to_fname.keys())

    name = None
    fname_to_desc = {}

    for tab in tables:
        for td in tab.findAll('td'):
            if name is not None:
                desc = td.text.strip()
                fname = name_to_fname[name]
                fname_to_desc[fname] = desc.replace('\r', '').replace('\n', '').replace('\t', '').replace('\xa0', '')
                name = None
            elif td.text.strip() in names_noext:
                name = td.text.strip()
    print(ub.urepr(fname_to_desc, nl=1))


def _build_test_image_table():
    """
    dev function for generating expected hashes

    Notes:
        No longer works, the URL is dead!
    """
    import os
    test_image_table = []
    for row in _TEST_IMAGES:
        fname = row['key']
        # fpath = grab_nitf_fpath(fname, safe=True)
        base = 'https://gwg.nga.mil/ntb/baseline/software/testfile/Nitfv2_1/'
        url = base + fname
        fpath = ub.grabdata(url, appname='kwgis/demodata/nitf')
        sha512 = ub.hash_file(fpath)[0:32]
        os.stat(fpath)
        new_row = ub.dict_union({
            'key': fname,
            'sha512': sha512,
            'size_bytes': os.stat(fpath).st_size,
        }, row)
        try:
            import kwimage
            kwimage.imread(fpath)
        except Exception:
            # Disable data that kwimage cant read
            new_row['enabled'] = False

        test_image_table.append(new_row)

    test_image_table = sorted(test_image_table, key=lambda x: x['size_bytes'])
    print('_TEST_IMAGES = {}'.format(ub.urepr(test_image_table, nl=1, sort=False)))


def _check_properties():
    from kwgis.gis.geotiff import geotiff_crs_info  # NOQA

    cid_map = ub.codeblock(
        '''
        QmfNRAZdn3VtEHQSzXRYdsFCZozMSVUuHv85yjyY6EPmka i_3001a.ntf
        QmSZkJKidBsWAUyPRFRbHFYxy1d4DBArhXD8utgavNASDp i_3004g.ntf
        QmVEfXgbQrHYVYbjiek5TEvxmvyWuLSLas5vnCmxBCJ5CG i_3008a.ntf
        QmPaDjuweukFaDiM44rxkHaSMwcvLu5gM6U8EY2XU6XzAF i_3015a.ntf
        QmfVBure4Jbmo7ESt5aotYbr5MnesWJxMFfRVLRoHqAfDd i_3018a.ntf
        QmfKK7NPfggEJAmVfv8VfFrzL6P8rpxTBUtX33YhHqmRF5 i_3025b.ntf
        QmSRGeqnkmMGfXFDZS22yWt3TAbxun533kNCWjuw6rqqpC i_3034c.ntf
        QmUHHD92Fecr1wviDihGUkEomVoF7XHSPYgK2vZN26Et9i i_3034f.ntf
        QmdNAsELfULbtXHTZhmfQ1svZVY24thttPpn4GNCZ9M2PE i_3041a.ntf
        QmZLE6WudN5ZRCwpMDnpK43BnERds3qNPanzqgCpuX1BZC i_3051e.ntf
        QmSkjd7DcqE34L8ThpAmpEPJNKa1ewSi2npQhrcjtreHmv i_3052a.ntf
        QmWCRDnbEkX7Nqju5vfUjetFTDek5GLHLjg2UwvHMsuffy i_3060a.ntf
        QmUFrYdz5xHrvMbcEWCpMurrqbXago7RK5U76EfnUrdniV i_3063f.ntf
        QmbxM3NB8iFJs5gQk2QJi4mFHbfFBcexpJEN6WsKq7Rq19 i_3068a.ntf
        QmUKSJRYKKDhsvZwdFgPnHLwzbEB22TFwwsoGxwCtB1Wne i_3076a.ntf
        QmXhiB3uFeNHT91PFSbSNEV3zbDrkGKr9FTGA3GeVrMbfU i_3090m.ntf
        QmYcA5UuwpMu6C7T92tLMMpmkMR69fzv3BdNt7y5KfDnTk i_3090u.ntf
        QmRSrjcm75KiCw5zjJFvWGJXSKaLLvRNpj7y8Bqd67uBnk i_3113g.ntf
        QmcAakF7RN533TiU8qnhGSUoKZuScCDXZVLeJmdcq7C9PG i_3114e.ntf
        QmYw4wppfG4uARMcaNrfrCCwrnTQmBb7cpW7VNZFRM7sXF i_3117ax.ntf
        QmQrnGbJ8XbUWipiw582WHbvpPqF2gLZsHj2tvRFgudC7H i_3128b.ntf
        QmNhrFM4SYsqUk333MCdTgzzwN5j1nvmCAu2Kjmq3ohEDm i_3201c.ntf
        QmeKe7twmNEq5iFTF6RZegVtuLBnm2GaVSnfLdbtLRkanh i_3228c.ntf
        QmYwNVAaC9bZNCN6T5Qjojjsub2C2sVWy7y5RD3ah7VtA2 i_3228e.ntf
        QmWqspPteMi38xHoohYLAhSqbcdNepWNNvS4f711uzicp8 i_3301a.ntf
        QmYYwyHfQnx4jdpsXuwQU1jTtHVWeaLsCq7ftFP6PkLP7H i_3301c.ntf
        QmSjNfgd2iue4ZoPbhzoTcp7isVK1aVBcDbPqYFsdRwTxg i_3301h.ntf
        ''')

    name_to_cid = {}
    for line in cid_map.split('\n'):
        line = line.strip()
        if line:
            cid, name = line.split(' ')
            name_to_cid[name] = cid

    infos = []
    for row in _TEST_IMAGES:
        if 'CID' not in row:
            row['CID'] = name_to_cid[row['key']]
        if row.get('enabled', True):
            print('----')
            fname = row['key']
            fpath = grab_nitf_fpath(fname, safe=True)
            out = ub.cmd('gdalinfo {}'.format(fpath), verbose=3)  # NOQA

            # if 'rpc' in out['out'].lower():
            #     break
            try:
                info = geotiff_crs_info(fpath)
                infos.append(info)
            except Exception as ex:
                print('ex = {!r}'.format(ex))
            else:
                row['has_crs'] = True
                row['has_rpc'] = info['is_rpc']
                print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
                _ = ub.cmd('gdalinfo {}'.format(fpath), verbose=3)
    [x['wld_crs_type'] for x in infos]
    print('_TEST_IMAGES = {}'.format(ub.urepr(_TEST_IMAGES, nl=1, sort=False)))



import ubelt as ub


def sentinel2_grid():
    """
    Grabs sentinel2 grid information from the web (if needed)

    Returns:
        geopandas.GeoDataFrame:
            A data frame where each row contains the tile name and the
            reversed-WGS84 (GeoJSON style) coordinates in
            shapely.geometry.Polygon format.

    References:
        https://gisgeography.com/arcgis-shapefile-files-types-extensions/
        https://github.com/justinelliotmeyers/Sentinel-2-Shapefile-Index

    CommandLine:
        xdoctest -m /home/joncrall/code/watch/kwgis/gis/sensors/sentinel2.py

    Example:
        >>> # xdoctest: +SKIP("slow, not currently used")
        >>> from kwgis.gis.sensors.sentinel2 import *  # NOQA
        >>> s2_tiles = sentinel2_grid()
        >>> assert s2_tiles.crs.name == 'WGS 84'
        >>> # Print out the first 5 tile rows
        >>> # print(s2_tiles.iloc[0:5])  # xdoctest: +IGNORE_WANT
        >>> '''
        >>>     Name                                           geometry
        >>> 0  01CCV  POLYGON Z ((180.00000 -73.05974 0.00000, 176.8...
        >>> 1  01CCV  POLYGON Z ((-180.00000 -72.07333 0.00000, -179...
        >>> 2  01CDH  POLYGON Z ((180.00000 -83.80855 0.00000, 174.7...
        >>> 3  01CDH  POLYGON Z ((-180.00000 -82.82590 0.00000, -176...
        >>> 4  01CDJ  POLYGON Z ((180.00000 -82.91344 0.00000, 175.7...
        >>> '''
        >>> #
        >>> # Demo how to convert each polygon into its UTM zone
        >>> import kwimage
        >>> from kwgis.utils import util_gis
        >>> import pyproj
        >>> utm_codes = []
        >>> # Only take some of the tiles for test speed
        >>> s2_tiles = s2_tiles.iloc[0:100]
        >>> for poly in ub.ProgIter(s2_tiles.geometry, desc='find utm'):
        >>>     lon = poly.centroid.x
        >>>     lat = poly.centroid.y
        >>>     utm_epsg = util_gis.utm_epsg_from_latlon(lat, lon)
        >>>     utm_codes.append(utm_epsg)
        >>> s2_tiles['utm_epsg'] = utm_codes
        >>> # Group all tiles within the same zone together
        >>> groups = dict(list(s2_tiles.groupby('utm_epsg')))
        >>> utm_groups = {}
        >>> for utm_epsg in ub.ProgIter(groups, desc='proj to utm'):
        >>>     utm_crs = pyproj.CRS.from_epsg(utm_epsg)
        >>>     # convert group into its utm coordinates
        >>>     utm_group = groups[utm_epsg].to_crs(utm_crs)
        >>>     utm_groups[utm_epsg] = utm_group
        >>> # Measure the area of each UTM Polygon
        >>> all_utm_areas = []
        >>> for utm_epsg, utm_group in utm_groups.items():
        >>>     all_utm_areas.extend([s.area for s in utm_group.geometry])
        >>> # Print out statistics about the chosen S2-tile UTM areas
        >>> import kwarray
        >>> area_stats = kwarray.stats_dict(all_utm_areas)
        >>> print('area_stats = {}'.format(ub.urepr(
        >>>     area_stats, nl=1, precision=0, align=':')))

        ...
        area_stats = {
            'mean' : 8463072768,
            'std'  : 3774869504,
            'min'  : 65260336,
            'max'  : 12056040448,
            ...
            'shape': (100,),
        }
    """
    from os.path import join
    import geopandas as gpd
    base_url = 'https://raw.githubusercontent.com/justinelliotmeyers/Sentinel-2-Shapefile-Index/9c553c340ee0b04f4d0c54c6f4d6c04504067dc6'
    items = {
        # Contains main geometry information
        'shp': {
            'fname': 'sentinel_2_index_shapefile.shp',
            'hash_prefix': '5343d73fa84e81cdcef03da86f4d21087f52cf41e4776aab1'},

        # contains name information
        'dbf': {
            'fname': 'sentinel_2_index_shapefile.dbf',
            'hash_prefix': 'ca65ffe251c0a1c4c0a27bf9aa915c475a56a4d6fb7fa465b'},

        # Contains CRS projection information
        'prj': {
            'fname': 'sentinel_2_index_shapefile.prj',
            'hash_prefix': 'b3ddd6fbfb5cca02d69acf67f986a7d1bd0400d72db141b61'},

        # Speeds up spatial queries
        'sbn': {
            'fname': 'sentinel_2_index_shapefile.sbn',
            'hash_prefix': 'f0f0118be76b3c53f44613f85d68c02d0b16f019b0c7630d3'},

        # Speeds up spatial queries
        'sbx': {
            'fname': 'sentinel_2_index_shapefile.sbx',
            'hash_prefix': 'ffe9c1dd8b55af0bc41f6cb354d4e8d107133b2d4e6f40e53'},

        # Required for spatial information?
        'shx': {
            'fname': 'sentinel_2_index_shapefile.shx',
            'hash_prefix': '4574fe1fe1be31e212b9c2ce0b09a162ccd518f67ccd455aa'},
    }
    for item in items.values():
        url = join(base_url, item['fname'])
        fpath = ub.grabdata(url, appname='kwgis', hash_prefix=item['hash_prefix'])
        item['fpath'] = fpath
    fpath = items['shp']['fpath']
    s2_tiles = gpd.read_file(fpath)
    return s2_tiles


def s2_grid_tiles_for_geometry(geometry):
    """
    Takes in a GeoJSON geometry and returns MGRS tile names that
    intersect it, ordered by intersection area (descending)

    Returns:
        List of strings:
            List of MGRS tile names

    Example:
        >>> # xdoctest: +SKIP("slow, not currently used")
        >>> from kwgis.gis.sensors.sentinel2 import *  # NOQA
        >>> from shapely.geometry import shape
        >>> geometry = shape({
        >>>  "type": "Polygon",
        >>>  "coordinates": [
        >>>     [
        >>>       [
        >>>         126.84333152242021,
        >>>         38.51998197874137
        >>>       ],
        >>>       [
        >>>         129.5107776938883,
        >>>         38.53873352863998
        >>>       ],
        >>>       [
        >>>         129.49645974534624,
        >>>         36.40500129696213
        >>>       ],
        >>>       [
        >>>         126.90373488362242,
        >>>         36.38763332541521
        >>>       ],
        >>>       [
        >>>         126.84333152242021,
        >>>         38.51998197874137
        >>>       ]
        >>>     ]
        >>>   ]
        >>>  })
        >>> assert(s2_grid_tiles_for_geometry(geometry) ==
        >>>        ['52SDG', '52SCG', '52SDH', '52SDF', '52SCH', '52SCF', '52SEG', '52SEH', '52SEF'])
    """

    s2_tiles = sentinel2_grid()

    # TODO: Convert geometries to UTM for more appropriate 'area'
    # computation

    s2_tiles['intersection_area'] =\
        s2_tiles.geometry.intersection(geometry).area

    return s2_tiles[s2_tiles['intersection_area'] > 0].sort_values(
        'intersection_area', ascending=False)['Name'].tolist()



"""
Tools to work with geotiff metadata.
"""
import numpy as np
import ubelt as ub
from kwgis.gis import spatial_reference as watch_crs
# from kwgis.utils.util_bands import LANDSAT7
from kwgis.utils import util_gis
from kwgis.utils.util_bands import SENTINEL2, LANDSAT8
import parse
from os.path import basename, isfile
from dateutil.parser import isoparse
from kwgis import exceptions


def geotiff_metadata(gpath, elevation='gtop30', strict=False,
                     supress_warnings=False):
    """
    Extract all relevant metadata we know how to extract.

    Args:
        gpath (str): path to the geotiff of interest
        elevation (str): method for extracting elevation data.

    Example:
        >>> # xdoctest: +REQUIRES(env:SLOW_DOCTEST)
        >>> # xdoctest: +REQUIRES(--network)
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> url = ('http://storage.googleapis.com/gcp-public-data-landsat/'
        ...        'LC08/01/044/034/LC08_L1GT_044034_20130330_20170310_01_T2/'
        ...        'LC08_L1GT_044034_20130330_20170310_01_T2_B11.TIF')
        >>> gpath = ub.grabdata(url, appname='kwgis')
        >>> info = geotiff_metadata(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))

        >>> import ubelt as ub
        >>> url = ('http://storage.googleapis.com/gcp-public-data-landsat/'
        ...        'LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/'
        ...        'LC08_L1TP_037029_20130602_20170310_01_T1_B2.TIF')
        >>> gpath = ub.grabdata(url, appname='kwgis')

        >>> info = geotiff_metadata(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))


    Ignore:
        source secrets/secrets
        export AWS_PROFILE=iarpa

        export XDEV_PROFILE=1
        pyblock "
        gpath = '/vsis3/smart-data-accenture/ta-1/ta1-wv-acc/52/S/DG/2020/4/13/20APR13020407-M1BS-014418171010_01_P004_ACC/20APR13020407-M1BS-014418171010_01_P004_ACC_B05.tif'
        from kwgis.gis.geotiff import geotiff_metadata
        info = geotiff_metadata(gpath)
        print(info)
        "
    """
    from kwgis.utils import util_gdal
    infos = {}
    ref = util_gdal.GdalDataset.open(gpath, 'r', virtual_retries=3)

    if supress_warnings:
        context = util_gdal.GdalSupressWarnings()
    else:
        import contextlib
        context = contextlib.nullcontext()

    infos['fname'] = geotiff_filepath_info(gpath)
    try:
        # TODO: we probably shouldn't suppress warnings here, remove once we
        # figure out why we are getting the current ones.
        with context:
            infos['crs'] = geotiff_crs_info(ref, elevation=elevation)
    except exceptions.GeoMetadataNotFound as ex:
        if strict:
            raise
        infos['crs'] = {'crs_error': str(ex)}
    infos['header'] = geotiff_header_info(ref)

    # Combine sensor candidates
    sensor_candidates = list(ub.flatten([
        v.get('sensor_candidates', []) for v in infos.values()]))

    info = ub.dict_union(*infos.values())
    info['sensor_candidates'] = sensor_candidates
    return info


def geotiff_header_info(gpath_or_ref):
    """
    Extract relevant metadata information from a geotiff header.

    Example:
        >>> # xdoctest: +REQUIRES(env:SLOW_DOCTEST)
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> from kwgis.demo.dummy_demodata import dummy_rpc_geotiff_fpath
        >>> from kwgis.demo.landsat_demodata import grab_landsat_product
        >>> gpath_or_ref = gpath = dummy_rpc_geotiff_fpath()
        >>> info = geotiff_header_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))
        >>> gpath_or_ref = gpath = grab_landsat_product()['bands'][0]
        >>> info = geotiff_header_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))
    """
    from kwgis.utils import util_gdal
    ref = util_gdal.GdalDataset.coerce(gpath_or_ref)
    keys_of_interest = [
        'NITF_CSEXRA_MAX_GSD',
        'NITF_PIAIMC_MEANGSD',
        'NITF_ISORCE',
        'NITF_PIAIMC_SENSNAME',
        'NITF_USE00A_MEAN_GSD',
        'NITF_CSEXRA_GEO_MEAN_GSD',
    ]
    sensor_candidates = []
    ignore_domains = {
        'xml:TRE', 'NITF_METADATA', 'DERIVED_SUBDATASETS', 'IMAGE_STRUCTURE',
        'TRE'}
    all_domain_img_meta = {}
    for ns in ref.GetMetadataDomainList():
        if ns in ignore_domains:
            continue
        meta = ref.GetMetadata(ns)
        if 0:
            print('meta = {}'.format(ub.urepr(meta, nl=1)))
        for key, value in ub.dict_isect(meta, keys_of_interest).items():
            all_domain_img_meta['{}.{}'.format(ns, key)] = value
    img_info = {}
    img_info['img_meta' ] = all_domain_img_meta
    img_info['num_bands'] = ref.RasterCount
    img_info['sensor_candidates'] = sensor_candidates

    band_metas = []
    for i in range(1, ref.RasterCount + 1):
        band = ref.GetRasterBand(i)
        # band.ComputeBandStats()
        if 0:
            band.GetStatistics(0, 1)
        band_meta_domains = band.GetMetadataDomainList() or []
        all_domain_band_meta = {}
        all_domain_band_meta['nodata'] = band.GetNoDataValue()
        for ns in band_meta_domains:
            band_meta = band.GetMetadata(ns)
            # band_meta = ub.dict_diff(band_meta, {'COMPRESSION'})
            for key, value in band_meta.items():
                all_domain_band_meta['{}.{}'.format(ns, key)] = value
        band_metas.append(all_domain_band_meta)

    if band_metas and any(band_metas):
        img_info['band_metas'] = band_metas

    if '.NITF_PIAIMC_SENSNAME' in all_domain_img_meta:
        sensor_candidates.append(all_domain_img_meta['.NITF_PIAIMC_SENSNAME'])
    return img_info


def geotiff_crs_info(gpath_or_ref, force_affine=False,
                     elevation='gtop30', verbose=0):
    """
    Use GDAL to extract coordinate system information about a geo_tiff.

    Builds transformations between pixel, geotiff-world, utm, and wgs84 spaces

    Args:
        gpath (str): path to the image file
        force_affine (bool): if True ignores RPC information
        elevation (str): method used to determine the elevation when RPC
            information is used. Available options are:
                * "open-elevation"
                * "gtop30"

    Example:
        >>> # xdoctest: +REQUIRES(env:SLOW_DOCTEST)
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> from kwgis.demo.dummy_demodata import dummy_rpc_geotiff_fpath
        >>> gpath_or_ref = gpath = dummy_rpc_geotiff_fpath()
        >>> info = geotiff_crs_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
        >>> assert info['is_rpc']
        >>> assert info['img_shape'] == (2000, 2000)

        >>> # xdoctest: +REQUIRES(--network)
        >>> gpath_or_ref = gpath = ub.grabdata(
        >>>     'https://download.osgeo.org/geotiff/samples/gdal_eg/cea.tif',
        >>>     appname='kwgis/demodata', hash_prefix='10a2ebcdcd95582')
        >>> info = geotiff_crs_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
        >>> assert not info['is_rpc']
        >>> assert info['img_shape'] == (515, 514)

        >>> # The public gateways seem to be too slow in serving the content
        >>> # xdoctest: +REQUIRES(--ipfs)
        >>> from kwgis.demo.nitf_demodata import grab_nitf_fpath
        >>> gpath_or_ref = gpath = grab_nitf_fpath('i_3004g.ntf')
        >>> info = geotiff_crs_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
        >>> assert not info['is_rpc']
        >>> assert info['img_shape'] == (512, 512)

        tf = info['wgs84_to_wld']
    """
    import affine
    import kwimage
    from kwgis.utils import util_gdal
    gdal = util_gdal.import_gdal()
    osr = util_gdal.import_osr()

    info = {}
    ref = util_gdal.GdalDataset.coerce(gpath_or_ref)

    if 0:
        # TODO: is it more efficient to use a gdal info call?  Do we get all
        # the information we need from it so we can serialize the data?
        json_info = gdal.Info(ref, format='json')
        json_info['coordinateSystem']
        aff_geo_transform = json_info['geoTransform']

        aff_wld_crs = osr.SpatialReference()
        aff_wld_crs.ImportFromWkt(json_info['coordinateSystem']['wkt'])
        # Not sure about how to import this info best
        if json_info['coordinateSystem']['dataAxisToSRSAxisMapping'] == [1, 2]:
            aff_wld_crs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        else:
            aff_wld_crs.SetAxisMappingStrategy(osr.OAMS_AUTHORITY_COMPLIANT)

    # tags = ref.GetMetadataDomainList()  # 7.5% of the execution time
    rpc_info = ref.GetMetadata(domain='RPC')  # 5% of execution time

    # if 'RPC' in tags:
    if len(rpc_info):
        rpc_info = ref.GetMetadata(domain='RPC')
        rpc_transform = watch_crs.RPCTransform.from_gdal(
            rpc_info, elevation=elevation)
        approx_elevation = rpc_transform._default_elevation()
    else:
        rpc_transform = None
        approx_elevation = None

    # TODO: understand the conditions for when these will not be populated
    # will proj always exist if wld_crs exists?
    if not hasattr(ref, 'GetSpatialRef'):
        import warnings
        warnings.warn('ref has no attribute GetSpatialRef, gdal version issue?')
        raise AssertionError('ref has no attribute GetSpatialRef, gdal version issue?')
        aff_wld_crs = None
    else:
        aff_wld_crs = ref.GetSpatialRef()  # 20% of the execution time
    aff_proj = ref.GetProjection()

    gcps = ref.GetGCPs()
    aff_gcp_wld_crs = ref.GetGCPSpatialRef()

    if gcps is not None:
        gcp_wld_coords = kwimage.Coords(np.array([[p.GCPX, p.GCPY] for p in gcps]))
        gcp_pxl_coords = kwimage.Coords(np.array([[p.GCPPixel, p.GCPLine] for p in gcps]))
    else:
        gcp_wld_coords = None
        gcp_pxl_coords = None

    if aff_wld_crs is None or aff_proj == '':
        aff_wld_crs = aff_gcp_wld_crs
        aff_wld_crs_type = 'gcp'  # mark the aff_wld crs as coming from the gcp
        if len(gcps) == 0 or aff_wld_crs is None:
            if rpc_transform is not None:
                # raise AssertionError('I dont think this should be possible')
                # Oh but it is, tests hit it.
                aff_wld_crs = osr.SpatialReference()
                aff_wld_crs.ImportFromEPSG(4326)  # 4326 is the EPSG id WGS84 of lat/lon crs
                # Set traditional because our rpc transform returns x,y
                aff_wld_crs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
                # Not 100% sure this is correct to do
                aff_wld_crs_type = 'assume-rpc-wgs84-reverse'
                aff_geo_transform = (0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
                if force_affine:
                    raise exceptions.GeoMetadataNotFound(
                        'cant force affine without dataset or gcp ref')
            else:
                raise exceptions.GeoMetadataNotFound('no dataset or gcps refs or rpc')
        else:
            # gcp_ids = [p.Id for p in gcps]
            aff_geo_transform = gdal.GCPsToGeoTransform(gcps)
    else:
        if aff_wld_crs is None:
            aff_wld_crs = osr.SpatialReference()
            aff_wld_crs.ImportFromWkt(aff_proj)
            aff_wld_crs_type = 'unknown?affine-projection-maybe?'
            raise AssertionError('can this ever happen?')
        else:
            # mark the aff_wld crs as coming from the dataset
            # is there a better name for this?
            aff_wld_crs_type = 'dataset'
        aff_geo_transform = ref.GetGeoTransform()

    if True:
        # TODO: do we need this info? If not, delete.
        # wld_crs = ref.GetSpatialRef()
        if aff_wld_crs_type == 'dataset':
            info['SpatialRefLinearUnits'] = aff_wld_crs.GetLinearUnits()
            info['SpatialRefLinearUnitsName'] = aff_wld_crs.GetLinearUnitsName()
            gt = aff_geo_transform
            pixelSizeX =  gt[1]
            pixelSizeY = -gt[5]
            info['GeoTransformGSD'] = (pixelSizeX, pixelSizeY)

        elif aff_wld_crs_type == 'gcp':

            aff_wld_crs = ref.GetGCPSpatialRef()
            gcps = ref.GetGCPs()
            if len(gcps) == 0 or aff_wld_crs is None:
                raise Exception('no gcps')
            gt = gdal.GCPsToGeoTransform(gcps)
            gt = ref.GetGeoTransform()
            pixelSizeX =  gt[1]
            pixelSizeY = -gt[5]
            info['GCPSpatialRefLinearUnits'] = aff_wld_crs.GetLinearUnits()
            info['GCPSpatialRefLinearUnitsName'] = aff_wld_crs.GetLinearUnitsName()
            info['GCPGeoTransformGSD'] = (pixelSizeX, pixelSizeY)

    aff = affine.Affine.from_gdal(*aff_geo_transform)
    aff_wld_from_pxl = np.vstack([np.array(aff.column_vectors).T, [0, 0, 1]])
    aff_pxl_from_wld = np.linalg.inv(aff_wld_from_pxl)

    is_rpc = not (force_affine or rpc_transform is None)

    if is_rpc:
        # If we are using RPC, we will ignore the "affine world CRS"
        # and define the standard one the RPC will warp into.
        # RPC is always wrt to WGS84.
        # TODO: Do we need to port an axis mapping strategy here?

        # FIXME: Our RPC transforms are currently using traditional
        # axis ordering. This is not compliant.

        wld_axis_mapping_int = osr.OAMS_TRADITIONAL_GIS_ORDER
        wld_axis_mapping = axis_mapping_int_to_text(wld_axis_mapping_int)
        wld_crs_type = 'hard-coded-rpc-crs'
        wld_crs = osr.SpatialReference()
        wld_crs.ImportFromEPSG(4326)  # 4326 is the EPSG id WGS84 of lat/lon crs
        wld_crs.SetAxisMappingStrategy(wld_axis_mapping_int)

        # aff_wld_axis_mapping_int = aff_wld_crs.GetAxisMappingStrategy()
        # wld_crs.SetAxisMappingStrategy(aff_wld_axis_mapping_int)
        # wld_axis_mapping_int = wld_crs.GetAxisMappingStrategy()
    else:
        wld_crs = aff_wld_crs
        wld_crs_type = aff_wld_crs
        wld_axis_mapping_int = aff_wld_crs.GetAxisMappingStrategy()
        wld_axis_mapping = axis_mapping_int_to_text(wld_axis_mapping_int)

    wgs84_crs = osr.SpatialReference()
    wgs84_crs.ImportFromEPSG(4326)  # 4326 is the EPSG id WGS84 of lat/lon crs
    wgs84_axis_mapping_int = wgs84_crs.GetAxisMappingStrategy()
    wgs84_axis_mapping = axis_mapping_int_to_text(wgs84_axis_mapping_int)

    wgs84_from_wld = osr.CoordinateTransformation(wld_crs, wgs84_crs)
    wld_from_wgs84 = osr.CoordinateTransformation(wgs84_crs, wld_crs)  # 10% of the execution time

    if is_rpc:
        pxl_from_wld = rpc_transform.make_warp_pixel_from_world()
        wld_from_pxl = rpc_transform.make_warp_world_from_pixel()
    else:
        wld_from_pxl = aff_wld_from_pxl
        pxl_from_wld = aff_pxl_from_wld

    shape = (ref.RasterYSize, ref.RasterXSize)

    # warp pixel corner points to lat / lon
    # Note, CCW order for conversion to polys
    xy_corners = np.array([
        [0, 0],
        [0, ref.RasterYSize],
        [ref.RasterXSize, ref.RasterYSize],
        [ref.RasterXSize, 0],
    ])
    pxl_corners = kwimage.Coords(xy_corners)
    wld_corners = pxl_corners.warp(wld_from_pxl)  # 10% of the execution time

    wgs84_corners = wld_corners.warp(wgs84_from_wld)
    min_lat, min_lon = wgs84_corners.data[:, 0:2].min(axis=0)
    max_lat, max_lon = wgs84_corners.data[:, 0:2].max(axis=0)

    # lat1 = wgs84_corners.data[:, 0].min()
    # lat2 = wgs84_corners.data[:, 0].max()
    # lon1 = wgs84_corners.data[:, 1].min()
    # lon2 = wgs84_corners.data[:, 1].max()

    # min_lon, max_lon = sorted([lon1, lon2])
    # min_lat, max_lat = sorted([lat1, lat2])

    assert util_gis.check_latlons(
        wgs84_corners.data[:, 0], wgs84_corners.data[:, 1]), (
            'bad WGS84 coordinates'
        )

    WITH_UTM_INFO = True
    if WITH_UTM_INFO:
        epsg_int = util_gis.utm_epsg_from_latlon(min_lat, min_lon)
        utm_crs = osr.SpatialReference()
        utm_crs.ImportFromEPSG(epsg_int)
        utm_axis_mapping_int = utm_crs.GetAxisMappingStrategy()
        utm_axis_mapping = axis_mapping_int_to_text(utm_axis_mapping_int)

        utm_from_wld = osr.CoordinateTransformation(wld_crs, utm_crs)  # 4% time
        # wld_from_utm = osr.CoordinateTransformation(utm_crs, wld_crs)
        utm_corners = wld_corners.warp(utm_from_wld)  # 2% time

        min_utm = utm_corners.data.min(axis=0)
        max_utm = utm_corners.data.max(axis=0)
        meter_extent = max_utm - min_utm
        pxl_extent = np.array([ref.RasterXSize, ref.RasterYSize])
        meter_per_pxl = meter_extent / pxl_extent
    else:
        meter_per_pxl = None
        meter_extent = None
        utm_from_wld = None
        # wld_from_utm = None
        utm_crs = None
        utm_axis_mapping = None

    if verbose:
        print('pxl_corners = {!r}'.format(pxl_corners))
        print('wld_corners = {!r}'.format(wld_corners))
        print('wgs84_corners = {!r}'.format(wgs84_corners))
        print('LAT: {} - {}'.format(min_lat, max_lat))
        print('LON: {} - {}'.format(min_lon, max_lon))

    bbox_geos = [min_lat, min_lon, max_lat, max_lon]

    if 1:
        # Is this a reasonable thing to do?
        utm_wkt = utm_crs.ExportToWkt()
        wld_wkt = wld_crs.ExportToWkt()
        wgs84_wkt = wgs84_crs.ExportToWkt()

        utm_crs_info = {
            'auth': memo_auth_from_wkt(utm_wkt),
            'axis_mapping': utm_axis_mapping,
        }

        wld_crs_info = {
            'auth': memo_auth_from_wkt(wld_wkt),
            'axis_mapping': wld_axis_mapping,
            'type': wld_crs_type,
        }

        wgs84_crs_info = {
            'auth': memo_auth_from_wkt(wgs84_wkt),
            'axis_mapping': wgs84_axis_mapping,
        }

        # Convert to the more general geos corners
        wgs84_crs_info = ub.dict_diff(wgs84_crs_info, {'type'})
        if wgs84_crs_info['axis_mapping'] == 'OAMS_AUTHORITY_COMPLIANT':
            geos_corners = kwimage.Polygon.coerce(
                wgs84_corners).swap_axes().to_geojson()
        else:
            geos_corners = kwimage.Polygon.coerce(
                wgs84_corners).to_geojson()
        geos_crs_info = {
            'axis_mapping': 'OAMS_TRADITIONAL_GIS_ORDER',
            'auth': ('EPSG', '4326')
        }
        geos_corners['properties'] = {'crs_info': geos_crs_info}

    info.update({
        'is_rpc': is_rpc,

        'meter_per_pxl': meter_per_pxl,
        'meter_extent': meter_extent,
        'approx_elevation': approx_elevation,

        'gcp_wld_coords': gcp_wld_coords,
        'gcp_pxl_coords': gcp_pxl_coords,

        'rpc_transform': rpc_transform,

        'pxl_corners': pxl_corners,
        'utm_corners': utm_corners,
        'wld_corners': wld_corners,
        'wgs84_corners': wgs84_corners,
        'geos_corners': geos_corners,

        # TODO: we changed the internal names to the "_from_" varaint but we
        # are keeping the "_to_" variant in this interface for now In the
        # future we should change things over to the "_from_" variant, which
        # is easier to reason about when chaining transforms.
        'pxl_to_wld': wld_from_pxl,
        'wgs84_to_wld': wld_from_wgs84,

        # 'utm_to_wld': wld_from_utm,  # unused, and 10% overhead


        'wld_to_pxl': pxl_from_wld,
        'wld_to_wgs84': wgs84_from_wld,
        'wld_to_utm': utm_from_wld,

        'utm_crs_info': utm_crs_info,
        'wld_crs_info': wld_crs_info,
        'wgs84_crs_info': wgs84_crs_info,

        'bbox_geos': bbox_geos,
        'img_shape': shape,
    })

    if info['utm_corners'] is not None:
        utm_box = kwimage.Polygon(exterior=info['utm_corners']).box()
        meter_w = float(utm_box.width)
        meter_h = float(utm_box.height)
        meter_hw = np.mean([meter_h , meter_w])
        pxl_hw = np.array(info['img_shape'])
        gsd = (meter_hw / pxl_hw).mean()
        minx, miny = info['utm_corners'].data.min(axis=0)
        maxx, maxy = info['utm_corners'].data.min(axis=0)
        info['approx_meter_gsd'] = gsd
    return info


def make_crs_info_object(osr_crs):
    """
    Args:
        osr_crs (osr.SpatialReference): an osr object from gdal

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> from kwgis.utils import util_gdal
        >>> osr = util_gdal.import_osr()
        >>> osr_crs = osr.SpatialReference()
        >>> osr_crs.ImportFromEPSG(4326)
        >>> crs_info = make_crs_info_object(osr_crs)
        >>> print('crs_info = {}'.format(ub.urepr(crs_info, nl=1)))
        >>> osr_crs = osr.SpatialReference()
        >>> osr_crs.ImportFromEPSG(4326)
        >>> osr_crs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        >>> crs_info = make_crs_info_object(osr_crs)
        >>> print('crs_info = {}'.format(ub.urepr(crs_info, nl=1)))
        >>> osr_crs.ImportFromEPSG(32744)
        >>> crs_info = make_crs_info_object(osr_crs)
        >>> print('crs_info = {}'.format(ub.urepr(crs_info, nl=1)))
    """
    wkt = osr_crs.ExportToWkt()
    auth = memo_auth_from_wkt(wkt)
    axis_mapping_int = osr_crs.GetAxisMappingStrategy()
    axis_mapping_text = axis_mapping_int_to_text(axis_mapping_int)
    crs_info = {
        'auth': auth,
        'axis_mapping': axis_mapping_text,
    }
    return crs_info


def axis_mapping_int_to_text(axis_mapping_int):
    """
    References:
        https://gdal.org/tutorials/osr_api_tut.html#crs-and-axis-order

    Notes:
        * OAMS_TRADITIONAL_GIS_ORDER means that for geographic CRS with
            lat/long order, the data will still be long/lat ordered.
            Similarly for a projected CRS with northing/easting order, the
            data will still be easting/northing ordered

        * OAMS_AUTHORITY_COMPLIANT means that the data axis will be
            identical to the CRS axis. This is the default value when
            instantiating OGRSpatialReference

        * OAMS_CUSTOM means that the data axis are customly defined with
            SetDataAxisToSRSAxisMapping
    """
    from kwgis.utils import util_gdal
    osr = util_gdal.import_osr()
    if axis_mapping_int == osr.OAMS_TRADITIONAL_GIS_ORDER:
        axis_mapping = 'OAMS_TRADITIONAL_GIS_ORDER'
    elif axis_mapping_int == osr.OAMS_AUTHORITY_COMPLIANT:
        axis_mapping = 'OAMS_AUTHORITY_COMPLIANT'
    elif axis_mapping_int == osr.OAMS_CUSTOM:
        axis_mapping = 'OAMS_CUSTOM'
    else:
        raise KeyError(axis_mapping_int)
    return axis_mapping


# def new_spatial_reference(axis_mapping='OAMS_AUTHORITY_COMPLIANT'):
#     """
#     Creates a new spatial reference

#     Args:
#         axis_mapping (int | str) : can be
#             OAMS_TRADITIONAL_GIS_ORDER, OAMS_AUTHORITY_COMPLIANT, or
#             OAMS_CUSTOM or the integer gdal code.

#     References:
#         https://gdal.org/tutorials/osr_api_tut.html#crs-and-axis-order
#     """
#     raise NotImplementedError
#     from osgeo import osr
#     if isinstance(axis_mapping, int):
#         axis_mapping_int = axis_mapping
#     else:
#         assert axis_mapping in {
#             'OAMS_TRADITIONAL_GIS_ORDER',
#             'OAMS_AUTHORITY_COMPLIANT',
#             'OAMS_CUSTOM',
#         }
#         axis_mapping_int = getattr(osr, axis_mapping)


@ub.memoize
def memo_auth_from_wkt(wkt):
    """
    This benchmarks as an expensive operation, memoize it.
    """
    from pyproj import CRS
    return CRS.from_wkt(wkt).to_authority(min_confidence=100)


class InvalidFormat(Exception):
    pass


def geotiff_filepath_info(gpath, fast=True):
    """
    Attempt to parse information out of a path to a geotiff file.

    Information provided here is purely heuristic. Generally filepath
    information is not robust.

    Several huerstics are currently implemented for:
        * Sentinel 2
        * Landsat-8
        * WorldView3

    See [S2_Name_2016]_ and [S3_Name]_.

    Args:
        gpath (str): a path to an image that uses a standard naming convention
            (may include subdirectories that contain relevant information) .

        fast (bool):
            if True stops when a hueristic matches well enough, otherwise tries
            multiple hueristics. Defaults to True.

    SeeAlso:
        * parse_landsat_product_id - specific to the landsat spec

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> inputs = [
        >>>     '18JUN18071532-S3DMR1C2.NTF',
        >>>     'LC08_L1GT_029030_20151209_20160131_01_RT',
        >>>     'LC08_L1TP_162043_20130401_20170505_01_T1_B8.tif',
        >>>     'LC08_L1GT_017037_20190303_20190309_01_T2_B8.TIF',
        >>>     '03AUG16WV031000016AUG03125811-P1BS-500827464010_01_P002_________AAE_0AAAAABPABB0.NTF',
        >>>     '13APR16WV030900016APR13131153-P1BS_R1C1-500659828010_01_P001____AAE_0AAAAABPABM0.NTF.NTF',
        >>>     '011777481010_01_P001_MUL/17SEP07021826-M1BS-011777481010_01_P001.TIF',
        >>>     '011778213010_01_P001_PAN/14FEB03022210-P1BS-011778213010_01_P001.NTF',
        >>>     'T17SMS_20151020T162042_TCI.jp2',
        >>>     'T39QXG_20160711T065622_TCI.jp2',
        >>>     'S2B_MSIL1C_20181219T022109_N0207_R003_T52SDG_20181219T051028.SAFE/GRANULE/L1C_T52SDG_A009324_20181219T022640/IMG_DATA/T52SDG_20181219T022109_TCI.jp2',
        >>>     'S2B_MSIL1C_20181219T022109_N0207_R003_T52SDG_20181219T051028/S2B_MSIL1C_20181219T022109_N0207_R003_T52SDG_20181219T051028.SAFE/GRANULE/L1C_T52SDG_A009324_20181219T022640/IMG_DATA/T52SDG_20181219T022109_B01.jp2',
        >>>     'S2A_MSIL1C_20151021T022702_N0204_R003_T52SDG_20151021T022701.SAFE/GRANULE/L1C_T52SDG_A001716_20151021T022701/IMG_DATA/T52SDG_20151021T022702_TCI.jp2',
        >>>     'LC08_L2SP_217076_20190107_20211008_02_T1_T23KPQ_B1.tif',
        >>>     'S2B_MSI_L2A_T23KPQ_20190114_20211008_SR_B01.tif',
        >>>     'S2A_MSI_L2A_T39RVK_20180803_20211102_SR_SOZ4.tif',
        >>> ]
        >>> gpath = inputs[-1]
        >>> print('gpath = {!r}'.format(gpath))
        >>> for gpath in inputs:
        >>>     print('----')
        >>>     info = geotiff_filepath_info(gpath)
        >>>     print('gpath = {}'.format(ub.urepr(gpath, nl=1)))
        >>>     print('info = {}'.format(ub.urepr(info, nl=2)))
        >>>     if len(info['sensor_candidates']) == 0:
        >>>         print(ub.color_text('NO HUERISTIC', 'red'))
        >>>     else:
        >>>         assert len(info['sensor_candidates']) == len(set(info['sensor_candidates']))
        >>>         assert info['filename_meta']

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> gpath = 'LC08_L1TP_037029_20130602_20170310_01_T1_B1.TIF'
        >>> info = geotiff_filepath_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))
        >>> assert info['filename_meta']['sensor_code'] == 'C'
        >>> assert info['filename_meta']['sat_code'] == '08'
        >>> assert info['filename_meta']['sat_code'] == '08'
        >>> assert info['filename_meta']['collection_category'] == 'T1'
        >>> assert info['filename_meta']['suffix'] == 'B1'

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> # Test extact info from real landsat product files
        >>> from kwgis.demo.landsat_demodata import grab_landsat_product  # NOQA
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> product = grab_landsat_product()
        >>> band_infos = [geotiff_filepath_info(gpath) for gpath in product['bands']]
        >>> meta_infos = [geotiff_filepath_info(gpath) for gpath in product['meta'].values()]
        >>> assert not any(d is None for d in band_infos)
        >>> assert not any(d is None for d in meta_infos)

    Ignore:
        >>> # TODO : demodata for a digital globe archive
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> import ubelt as ub
        >>> gpath = ub.expandpath('$HOME/remote/namek/data/dvc-repos/smart_watch_dvc/drop0/KR-Pyeongchang-WV/_assets/20170907_a_KRP_011777481_10_0/011777481010_01_003/011777481010_01/011777481010_01_P001_MUL/17SEP07021826-M1BS-011777481010_01_P001.TIF')
        >>> info = geotiff_filepath_info(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))

    Ignore:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> base_dpath = '/home/joncrall/data/grab_tiles_out/fels'
        >>> paths = sorted(walk_geotiff_products(base_dpath, with_product_dirs=1, with_loose_images=0))
        >>> infos = []
        >>> for path in paths:
        >>>     info = geotiff_filepath_info(path)
        >>>     info['basename'] = basename(path)
        >>>     print('info = {}'.format(ub.urepr(info, nl=1)))
        >>>     infos.append(info)

        >>> paths = sorted(walk_geotiff_products(base_dpath, with_product_dirs=0, with_loose_images=1))
        >>> infos = []
        >>> for path in paths:
        >>>     info = geotiff_filepath_info(path)
        >>>     info['basename'] = basename(path)
        >>>     print('info = {}'.format(ub.urepr(info, nl=1)))
        >>>     infos.append(info)
        >>> print('infos = {}'.format(ub.urepr(infos[0:10], nl=1)))
        >>> print('infos = {}'.format(ub.urepr(infos[-10:], nl=1)))

        >>> infos = []
        >>> for path in paths:
        >>>     info = geotiff_filepath_info(path)
        >>>     info['path'] = path
        >>>     infos.append(info)

    References:
        .. [S2_Name_2016] https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention
        .. [S3_Name] https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-altimetry/naming-conventions

    """

    base_ext = basename(gpath)
    base, *exts = base_ext.split('.')
    ext = '.'.join(exts)  # NOQA
    parts = gpath.split('/')

    info = {
        'sensor_candidates': [],
        'filename_meta': {},
    }
    sensor_candidates = info['sensor_candidates']
    meta = info['filename_meta']

    ls_meta = parse_landsat_product_id(base)
    if ls_meta is not None:
        sensor_cand = 'L' + ls_meta['sensor_code'] + ls_meta['sat_code']
        meta.update(ls_meta)
        meta['product_guess'] = 'landsat'
        meta['guess_heuristic'] = 'landsat_parse'
        sensor_candidates.append(sensor_cand)

    if not sensor_candidates or not fast:
        s2_meta = parse_sentinel2_product_id(parts)
        if s2_meta is not None:
            meta.update(s2_meta)
            sensor_candidates.append(meta['mission_id'])
            if meta.get('suffix', None) == 'TCI':
                sensor_candidates.append('S2-TrueColor')
            meta['product_guess'] = 'sentinel2'
            meta['guess_heuristic'] = 'sentinel2_parse'

    dg_bundle = None
    if not sensor_candidates or not fast:
        # WorldView3
        # TODO: find a reference for the spec
        # TODO fix date handling for eg 03JUL15WV020200015JUL03021500-P1BS-011777484010_01_P001.NTF or 15JUL03021500
        wv3_pat = '{date1:w}-{part1:w}-{date2:w}_{num:w}_{part2:w}'
        wv3_parser = _parser_lut(wv3_pat)
        prefix = base.split('____')[0]
        result = wv3_parser.parse(prefix)
        if result is not None:
            try:
                wv2_meta = {}
                wv2_meta['date1'] = result.named['date1']
                wv2_meta['part1'] = result.named['part1']
                wv2_meta['date2'] = result.named['date2']
                wv2_meta['num'] = result.named['num']
                wv2_meta['part2'] = result.named['part2']
                wv2_meta['product_guess'] = 'worldview'
                wv2_meta['guess_heuristic'] = 'WV_heuristic1'
                meta.update(wv2_meta)
            except InvalidFormat:
                pass
            else:
                meta.update(wv2_meta)
                sensor_candidates.append('WV03')

        # Add DG information if it exists
        from kwgis.gis import digital_globe as dg_parser
        try:
            # technically, this does read files, so its not all about the path
            dg_bundle = dg_parser.DigitalGlobeBundle.from_pointer(gpath)
            info['dg_info'] = dg_bundle.data
        except Exception:
            dg_bundle = None

        if dg_bundle is not None:
            for prod_meta in dg_bundle.data['product_metas']:
                sensor_candidates.append(prod_meta['sensorVehicle'])

    # TODO: handle landsat and sentinel2 bundles
    info['is_dg_bundle'] = dg_bundle is not None

    if 0:
        # This is slow, and I don't think it is ever hit

        def _is_rgb(gpath):
            # fallback for 'channels'
            # often, a gtiff is a TCI that was postprocessed in some way that destroys
            # the original naming convention
            #
            # this opens the image to check for that case as a fallback
            from kwgis.utils import util_gdal
            gdal = util_gdal.import_gdal()
            info = gdal.Info(gpath, format='json')
            if len(info['bands']) == 3:
                # TODO sometimes colorInterpretation is stripped, but it's still RGB
                # should this return True for any gpath with 3 bands?
                if [b['colorInterpretation'] for b in info['bands']] == ['Red', 'Green', 'Blue']:
                    return True
            return False

        if 'channels' not in meta:
            if isfile(gpath) and _is_rgb(gpath):
                meta['channels'] = 'r|g|b'

    return info


@ub.memoize
def _parser_lut(pattern):
    """
    Calling parse.parse is about 14x slower than creating a parser object
    once and then using it.
    """
    return parse.Parser(pattern)


def parse_sentinel2_product_id(parts):
    """
    Try to parse the Sentinel-2 pre-2016 and post-2016 safedir formats.

    Note that unlike parse_landsat_product_id, which expects a band file basename,
    this presently purports to plurally parse pieces of path postfixedly
    (it parses the whole path, backwards :))

    TODO extend this to parsing the granuledir and band file formats as well.
    For now, we just need all names to be minimally parseable, even if some info is incorrect.

    General plan is to check the old formats strictly first, and then check the new safedir loosely as a default

    Example:

        parts = ['S2B_MSI_L2A_T23KPQ_20190114_20211008_SR_B01.tif']
        parts = ['S2A_MSI_L2A_T39RVK_20180803_20211102_SR_SOZ4.tif']
        parse_sentinel2_product_id(parts)

    """

    def _dt(name):
        # expand to a named ISO 8601 datetime without separators, which is not supported by parse
        # example: 20190901T234135
        # {{ escapes {

        # trying to be too clever here...
        # return f'{{{name}.Y:04d}}{{{name}.M:02d}}{{{name}.D:02d}}T{{{name}.h:02d}}{{{name}.m:02d}}{{{name}.s:02d}}'
        # return f'{{{name}.date:08d}}T{{{name}.time:06d}}'
        return f'{{{name}:.15}}'

    # unfortunately parse() doesn't seem to support a format specifier for "string of length exactly n"
    # {name:n} is ">= n"
    # {name:.n} is "<= n"  <- going with this one as a better approximation of "exactly n"

    # this also EXCLUDES the trailing '.SAFE' for all safedirs, again because parse can't handle optional pieces

    s2_safedir_2015 = '{MMM:.3}_{CCCC:.4}_PRD_{MSIXXX:.6}_{ssss:.4}_' + _dt('creation') + '_R{OOO:03d}_V' + _dt('sensing_start') + '_' + _dt('sensing_end')
    s2_safedir_2015_parser = _parser_lut(s2_safedir_2015)

    # parse also doesn't support 'or'
    # A could instead be the Validity Start Time
    # T could instead be the Detector ID
    # but these are the more common (and useful) choices
    s2_granuledir_2015 = '{MMM:.3}_{CCCC:.4}_MSI_{YYY:.3}_{ZZ:.2}_{ssss:.4}_' + _dt('validity_start') + '_A{ffffff:06d}_T{xxxxx:.5}_N{xx:02d}.{yy:02d}'
    s2_granuledir_2015_parser = _parser_lut(s2_granuledir_2015)

    # Sentinel-2 2016+ filename pattern. See [S2_Name_2016]_
    # These filenames are often directories
    # MMM_MSIXXX_YYYYMMDDHHMMSS_Nxxyy_ROOO_Txxxxx_<Product Discriminator>.SAFE
    # SAFE = Standard Archive Format for Europe
    s2_name_2016 = '{MMM:.3}_{MSIXXX:.6}_{YYYYMMDDHHMMSS:.15}_{Nxxyy:.5}_{ROOO:.4}_{Txxxxx:.6}_{Discriminator}'
    s2_2016_parser = _parser_lut(s2_name_2016)

    # use util_bands for this
    s2_channel_alias = {
        band['name']: band['common_name'] for band in SENTINEL2 if 'common_name' in band
    }
    # ...except for TCI, which is not a true band, but often included anyway
    # and this channel code is more specific to kwcoco
    s2_channel_alias.update({'TCI': 'r|g|b'})

    meta = {}

    # TODO allow for parsing multiple parts once safedir, granuledir, and band file are all implemented
    for _part in reversed(parts):

        part = _part.split('.')[0]
        result = s2_safedir_2015_parser.parse(part)
        if result:
            s2_meta = {}
            try:
                mission_id = result.named['MMM']
                if mission_id not in {'S2A', 'S2B'}:
                    raise InvalidFormat
                s2_meta['mission_id'] = mission_id
                s2_meta['product_level'] = result.named['MSIXXX']
                s2_meta['sense_start_time'] = isoparse(result.named['sensing_start']
                                                       ).isoformat()
                s2_meta['relative_orbit_num'] = result.named['ROOO']
                s2_meta['discriminator'] = result.named['Discriminator']
                s2_meta['product_guess'] = 'sentinel2'
                s2_meta['guess_heuristic'] = 'S2_safedir_2015_format'
            except InvalidFormat:
                pass
            else:
                meta.update(s2_meta)
                break

        part = '.'.join(_part.split('.')[:2])
        result = s2_granuledir_2015_parser.parse(part)
        if result:
            s2_meta = {}
            try:
                mission_id = result.named['MMM']
                if mission_id not in {'S2A', 'S2B'}:
                    raise InvalidFormat
                s2_meta['mission_id'] = mission_id
                s2_meta['product_level'] = result.named['YYY']
                # TODO warning, this date can differ between the 2 in safedir and 1 in granuledir!
                # it's an open question which should be "canonical"
                s2_meta['sense_start_time'] = isoparse(result.named['validity_start']
                                                       ).isoformat()
                s2_meta['pdgs_num'] = f"N{result.named['xx']:02d}{result.named['yy']:02d}"
                s2_meta['absolute_orbit_num'] = result.named['ffffff']
                s2_meta['tile_number'] = 'T' + result.named['xxxxx']
                s2_meta['product_guess'] = 'sentinel2'
                s2_meta['guess_heuristic'] = 'S2_2016_format'
            except InvalidFormat:
                pass
            else:
                meta.update(s2_meta)
                break

        part = _part.split('.')[0]
        result = s2_2016_parser.parse(part)
        if result:
            s2_meta = {}
            try:
                mission_id = result.named['MMM']
                if mission_id not in {'S2A', 'S2B'}:
                    raise InvalidFormat
                s2_meta['mission_id'] = mission_id
                s2_meta['product_level'] = result.named['MSIXXX']
                s2_meta['sense_start_time'] = result.named['YYYYMMDDHHMMSS']
                s2_meta['pdgs_num'] = result.named['Nxxyy']
                s2_meta['relative_orbit_num'] = result.named['ROOO']
                s2_meta['tile_number'] = result.named['Txxxxx']
                s2_meta['discriminator'] = result.named['Discriminator']
                s2_meta['product_guess'] = 'sentinel2'
                s2_meta['guess_heuristic'] = 'S2_2016_format'
            except InvalidFormat:
                pass
            else:
                meta.update(s2_meta)
                break

    # Files ending in _TCI are true color images based on the Sentinel 2
    # Handbook I'm not sure what the standard for this format is I just know a
    # suffix of _TCI means true color image. ANd these are from sentinel2, I'm
    # guessing on the rest of the format.
    s2_format_guess1 = '{tile_number:.6}_{date:.15}_{band:.3}.{ext}'
    s2_parser = _parser_lut(s2_format_guess1)
    result = s2_parser.parse(parts[-1])
    if result:
        tile_number = result.named['tile_number']
        if len(tile_number) == 6 and tile_number.startswith('T'):
            if 'sense_start_time' in meta:
                assert meta['sense_start_time'] == result.named['date']
            # Changed from acquisition_date to match other S2 information
            meta['sense_start_time'] = result.named['date']
            band = result.named['band']
            # TODO: normalized consise channel code
            meta['tile_number'] = tile_number
            meta['suffix'] = band
            channels = s2_channel_alias.get(band, band)
            meta['channels'] = channels
            meta['product_guess'] = 'sentinel2'
            meta['guess_heuristic'] = 'S2_tile_date_band_format'
            if 'mission_id' not in meta:
                meta['mission_id'] = 'S2'

    # This is another guess based on a file that failed to ingest
    s2_format_guess2 = '{MMM:.3}_MSI_{XXX:.3}_{Txxxxx:.6}_{SENSE_YYYYMMDD:.8}_{PROC_YYYYMMDD:.8}_{correction_code}_{band}.{ext}'
    s2_parser2 = _parser_lut(s2_format_guess2)
    result = s2_parser2.parse(parts[-1])
    if result:
        s2_meta = {}
        mission_id = result.named['MMM']
        if mission_id not in {'S2A', 'S2B'}:
            raise InvalidFormat
        s2_meta['mission_id'] = mission_id
        s2_meta['product_level'] = result.named['XXX']
        s2_meta['tile_number'] = result.named['Txxxxx']
        s2_meta['sense_start_time'] = result.named['SENSE_YYYYMMDD']
        s2_meta['processing_date'] = result.named['PROC_YYYYMMDD']  # is this corret?
        s2_meta['correction_code'] = result.named['correction_code']
        s2_meta['product_guess'] = 'sentinel2'
        s2_meta['guess_heuristic'] = 's2_format_guess2'
        s2_meta['band'] = band = result.named['band']
        s2_meta['channels'] = s2_channel_alias.get(band, band)
        # l8_channel_alias = {
        #     band['name']: band['common_name'] for band in SENTINEL2 if 'common_name' in band
        # }

        meta.update(s2_meta)

    if meta:
        if 'sense_start_time' in meta:
            # Write data to a consistent key
            meta['date_captured'] = isoparse(meta['sense_start_time']).isoformat()
        return meta


def parse_landsat_product_id(product_id):
    """
    Extract information from a landsat produt id

    See [LanSatName]_, [LS_578]_, [ExampleLandSat]_, [LandSatSuffixFormat]_,
    [LandsatProcLevels]_, [LandsatL2Names]_, and [LandSatARDDocs]_.

    Args:
        product_id (str): this is typically the filename (without extension!)
            of a landsat product, as described in [LanSatName]_.

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> product_id = 'LC08_L1TP_037029_20130602_20170310_01_T1'
        >>> ls_meta = parse_landsat_product_id(product_id)

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> gpath = 'LC08_L1TP_037029_20130602_20170310_01_T1_B1'
        >>> info = parse_landsat_product_id(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))
        >>> assert info['sensor_code'] == 'C'
        >>> assert info['sat_code'] == '08'
        >>> assert info['sat_code'] == '08'
        >>> assert info['collection_category'] == 'T1'
        >>> assert info['suffix'] == 'B1'
        >>> assert info['band_num'] == 1


        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> gpath = 'LC08_CU_029005_20181208_20210503_02_QA_LINEAGE.TIF'
        >>> info = parse_landsat_product_id(gpath)
        >>> print('info = {}'.format(ub.urepr(info, nl=1)))

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> # Test on real landsat data
        >>> from kwgis.demo.landsat_demodata import grab_landsat_product  # NOQA
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> product = grab_landsat_product()
        >>> band_prodids = [ub.augpath(gpath, dpath='', ext='') for gpath in product['bands']]
        >>> band_infos = [parse_landsat_product_id(product_id) for product_id in band_prodids]
        >>> assert ub.allsame([ub.dict_diff(d, ['band_num', 'suffix', 'channels', 'band']) for d in band_infos])
        >>> meta_prodids = [ub.augpath(gpath, dpath='', ext='') for gpath in product['meta'].values()]
        >>> meta_infos = [parse_landsat_product_id(product_id) for product_id in meta_prodids]
        >>> assert ub.allsame([ub.dict_diff(d, ['suffix', 'channels', 'band']) for d in meta_infos])

    References:
        .. [LanSatName] https://www.usgs.gov/faqs/what-naming-convention-landsat-collections-level-1-scenes
        .. [LS_578] https://github.com/dgketchum/Landsat578#-1
        .. [ExampleLandSat]  https://console.cloud.google.com/storage/browser/gcp-public-data-landsat/LC08/01/044/034/LC08_L1GT_044034_20130330_20170310_01_T2?_ga=2.210779154.665659046.1615242530-37570621.1615242530
        .. [LandSatSuffixFormat] https://prd-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/atoms/files/LSDS-750_Landsat8_Level-0-Reformatted_DataFormatControlBook-v15.pdf (page 26 / 99)
        .. [LandsatProcLevels] https://www.usgs.gov/core-science-systems/nli/landsat/landsat-levels-processing
        .. [LandsatL2Names] https://www.usgs.gov/faqs/what-naming-convention-landsat-collection-2-level-1-and-level-2-scenes?qt-news_science_products=0#qt-news_science_products

        .. [LandSatARDDocs] https://d9-wret.s3.us-west-2.amazonaws.com/assets/palladium/production/s3fs-public/atoms/files/LSDS-1873_US-Landsat%20C1-ARD-DFCB-v7.pdf
    """
    # Landsat filename pattern. See [LanSatName]_
    # LXSS_LLLL_PPPRRR_YYYYMMDD_yyyymmdd_CC_TX
    #                   LXSS     _ LLLL_  PPPRRR _ YYYYMMDD _ yyyymmdd _ CC _ TX
    from dateutil.parser import isoparse
    landsat_pattern = 'L{X:1}{SS}_{LLLL}_{PPPRRR}_{YYYYMMDD}_{yyyymmdd}_{CC}_{TX}'
    landsat_parser = _parser_lut(landsat_pattern)
    result = landsat_parser.parse(product_id)
    if result:
        ls_sensor_code_to_text = {
            'C': 'OLI/TIRS',
            'O': 'OLI',
            'T': 'TIRS',
            'E': 'ETM+',
            # 'T': 'TM',  # ambiguous? That's in the spec
            'M': 'MSS',
        }

        correction_code_to_text = {
            'L1TP': 'Precision Terrain',
            'L1GT': 'Systematic Terrain',
            'L1GS': 'Systematic',
            'L2SP': 'Science Package',
            'L2SR': 'Surface Reflectance',
            'L2ST': 'Surface Temperature',  # this is a guess
            'CU': 'CU?',  # No idea what this is. Seen in quality ands.
        }

        # use util_bands for this
        # l7_channel_alias = {
        #     band['name']: band['common_name'] for band in LANDSAT7 if 'common_name' in band
        # }
        l8_channel_alias = {
            band['name']: band['common_name'] for band in LANDSAT8 if 'common_name' in band
        }

        # When accessing files from google API, there might be an additional
        # field specifying band information.
        trailing = result.named['TX'].split('_')
        tx = trailing[0]

        wrs = result.named['PPPRRR']
        sensor_code = result.named['X']
        sat_code = result.named['SS']

        correction_code = result.named['LLLL']

        ls_meta = {}
        ls_meta['sensor_text']: ls_sensor_code_to_text[sensor_code]
        ls_meta['correction_level_text'] = correction_code_to_text[correction_code]

        ls_meta['sensor_code'] = sensor_code
        ls_meta['sat_code'] = sat_code
        ls_meta['WRS_path'] = wrs[:3]
        ls_meta['WRS_row'] = wrs[3:]
        ls_meta['tile_number'] = wrs
        ls_meta['correction_level_code'] = correction_code
        ls_meta['acquisition_date'] = result.named['YYYYMMDD']
        ls_meta['processing_date'] = result.named['yyyymmdd']
        ls_meta['collection_number'] = result.named['CC']
        ls_meta['collection_category'] = tx

        # Common key
        ls_meta['date_captured'] = isoparse(ls_meta['acquisition_date']).isoformat()

        if len(trailing) > 1:
            suffix = '_'.join(trailing[1:])
            ls_meta['suffix'] = suffix

            name_suffix = suffix
            removable_exts = ['.tif', '.tiff']
            for s in removable_exts:
                if name_suffix.lower().endswith(s):
                    name_suffix = name_suffix[:-len(s)]

            ls_meta['band'] = name_suffix
            ls_meta['channels'] = l8_channel_alias.get(name_suffix, name_suffix)

            if name_suffix == 'ANC':
                ls_meta['is_ancillary'] = True
            elif name_suffix == 'MTA':
                ls_meta['is_metadata'] = True
            elif name_suffix == 'MD5':
                ls_meta['is_checksum'] = True
            else:
                # The suffix might represent something about band
                # information, we may parse it.
                # See [LandSatSuffixFormat]_.
                band_suffix_pat = 'B{band_num:d}'
                band_suffix_parser = _parser_lut(band_suffix_pat)
                band_result = band_suffix_parser.parse(name_suffix)
                if band_result is not None:
                    ls_meta['band_num'] = band_result.named['band_num']
        return ls_meta


# def normalize_sensor():
#     pass


def walk_geotiff_products(dpath, with_product_dirs=True,
                          with_loose_images=True, recursive=True):
    """
    Walks a file path and returns directories and files that look
    like standalone geotiff products.

    Args:
        dpath (str): directory to search

    Yields:
        str: paths of files or recognized geotiff product bundle directories

    Example:
        >>> # xdoctest: +REQUIRES(env:SLOW_DOCTEST)
        >>> # xdoctest: +REQUIRES(--network)
        >>> # Test on real landsat data
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> import kwgis
        >>> from os.path import dirname
        >>> product = kwgis.demo.landsat_demodata.grab_landsat_product()
        >>> dpath = dirname(dirname(ub.peek(product['bands'])))
        >>> print(list(walk_geotiff_products(dpath)))
        >>> print(list(walk_geotiff_products(dpath, with_product_dirs=False)))
    """
    import os
    from os.path import join
    # blocklist = set()
    GEOTIFF_EXTENSIONS = ('.vrt', '.tiff', '.tif', '.jp2')

    for r, ds, fs in os.walk(dpath):
        handled = []
        if with_product_dirs:
            for didx, dname in enumerate(ds):
                if dname.startswith('LE07'):
                    dpath = join(r, dname)
                    handled.append(didx)
                    yield dpath
                elif dname.startswith('LC08_'):
                    dpath = join(r, dname)
                    handled.append(didx)
                    yield dpath
                elif dname.startswith(('S2A_', 'S2B_')):
                    handled.append(didx)
                    dpath = join(r, dname)
                    yield dpath
                else:
                    pass
            for didx in reversed(handled):
                del ds[didx]

        if with_loose_images:
            for fname in fs:
                if fname.lower().endswith(GEOTIFF_EXTENSIONS):
                    fpath = join(r, fname)
                    yield fpath

        if not recursive:
            break



"""
Generate demodata for the metrics framework
"""
import click
import json
import kwarray
import ubelt as ub
from kwgis.demo.metrics_demo import demo_truth
from kwgis.demo.metrics_demo import demo_rendering
from kwgis.demo.metrics_demo import site_perterbing


def generate_demo_metrics_framework_data(
        roi="DR_R001",
        num_sites=3,
        num_observations=5,
        noise=1.0,
        p_observe=0.5,
        p_transition=0.15,
        seed=409060576688592,
        outdir=None,
        cache=True,
        **kwargs):
    """
    Example:
        >>> from kwgis.demo.metrics_demo.generate_demodata import *  # NOQA
        >>> demo_info = generate_demo_metrics_framework_data(
        >>>     num_sites=5, num_observations=10, noise=2, p_observe=0.5,
        >>>     p_transition=0.3, drop_noise=0.5, drop_limit=0.5)
        >>> print('demo_info = {}'.format(ub.urepr(demo_info, nl=1)))
        >>> # TODO: visualize

    Ignore:
        from kwgis.demo.metrics_demo.generate_demodata import *  # NOQA
        import xdev
        globals().update(xdev.get_func_kwargs(generate_demo_metrics_framework_data))
    """

    if outdir is None:
        demo_dpath = ub.Path.appdir("iarpa/smart/ta2/demodata")
    else:
        demo_dpath = ub.Path(outdir)

    demo_dpath.ensuredir()

    image_dpath = demo_dpath / "images"
    true_dpath = demo_dpath / "truth"
    pred_dpath = demo_dpath / "predictions"
    output_dpath = demo_dpath / "demo_output"
    cache_dpath = demo_dpath / "_cache"

    output_dpath.ensuredir()
    true_region_dpath = (true_dpath / "region_models").ensuredir()
    true_site_dpath = (true_dpath / "site_models").ensuredir()
    pred_site_dpath = (pred_dpath / "pred_site_models").ensuredir()

    performer_id = 'alice'
    region_id = roi

    truekw = {
        'region_id': region_id,
        'num_sites': num_sites,
        'num_observations': num_observations,
        'p_transition': p_transition,
        'p_observe': p_observe,
    }
    predkw = {
        'noise': noise,
        'performer_id': performer_id,
        **kwargs,
    }

    rng = kwarray.ensure_rng(seed)
    # TODO: add more parameters here
    depends = [
        rng,
        truekw,
        predkw,
        "v2",
    ]

    stamp = ub.CacheStamp("demodata", dpath=cache_dpath, depends=depends,
                          enabled=cache)
    if stamp.expired():
        cache_dpath.ensuredir()
        # Generate a random truth region
        region, sites, renderables = demo_truth.random_region_model(
            rng=rng, **truekw)

        # Generate predictions that are close to the truth, but not quite.
        # TODO: parametarize the amount of perterbation
        pred_sites = site_perterbing.perterb_site_model(sites, rng=rng,
                                                        **predkw)

        # Dump region file
        region_id = region["features"][0]["properties"]["region_id"]
        region_fpath = true_region_dpath / (region_id + ".geojson")
        region_fpath.write_text(json.dumps(region, indent="  "))
        print(f"wrote true region_fpath={region_fpath}")

        # Dump true site file
        for site in sites:
            site_id = site["features"][0]["properties"]["site_id"]
            # site_fpath = (true_site_dpath / region_id).ensuredir() / (site_id + '.geojson')
            site_fpath = true_site_dpath / (site_id + ".geojson")
            site_fpath.write_text(json.dumps(site, indent="  "))
            print(f"wrote site_fpath={site_fpath}")

        # Dump pred site file
        for site in pred_sites:
            site_id = site["features"][0]["properties"]["site_id"]
            site_fpath = pred_site_dpath / (site_id + ".geojson")
            site_fpath.write_text(json.dumps(site, indent="  "))
            print(f"wrote predicted site_fpath={site_fpath}")

        # Render toy images
        # The image path currently needs to in a very specific format
        region_img_dpath = (image_dpath.joinpath(*region_id.split('_')) / 'images' / 'a' / 'b').ensuredir()
        # TODO: also need to output the crops path as well for visualize_stack_slices.
        # region_img_dpath = (image_dpath / region_id).ensuredir()
        for renderable in ub.ProgIter(renderables, desc="render"):
            demo_rendering.render_toy_georeferenced_image(
                region_img_dpath, renderable, rng=rng)
        stamp.renew()

    demo_info = {
        "region_id": region_id,
        "true_site_dpath": true_site_dpath,
        "true_region_dpath": true_region_dpath,
        "pred_site_dpath": pred_site_dpath,
        "image_dpath": image_dpath,
        "output_dpath": output_dpath,
    }
    return demo_info


@click.command()
@click.option('--roi', default="DR_R001", help="name of the demo region interest")
@click.option('--num_sites', default=3, help="number of demo sites to generate")
@click.option('--num_observations', default=5, help="number of observations of the demo region")
@click.option('--noise', default=1.0, help="Noise level of perterbations")
@click.option('--p_observe', default=0.5, help="The probability of annotating an observation")
@click.option('--p_transition', default=0.5, help="The probability of a site transitions phase on each frame")
@click.option('--seed', default=409060576688592, help="RNG seed")
@click.option('--outdir', default=None, help="Where to write the data. Defaults to ~/.cache/iarpa/smart/ta2/demodata")
@click.option('--cache/--no-cache', default=True, help="enable/disable the cache")
def main(*args, **kwargs):
    import os
    demo_info = generate_demo_metrics_framework_data(*args, **kwargs)
    # Build command to tell the user how to test this data with the eval
    # framework.
    test_args = [
        'python', '-m', 'iarpa_smart_metrics.run_evaluation',
        '--roi', demo_info['region_id'],
        '--gt_dir', os.fspath(demo_info['true_site_dpath']),
        '--rm_dir', os.fspath(demo_info['true_region_dpath']),
        '--sm_dir', os.fspath(demo_info['pred_site_dpath']),
        '--image_dir', os.fspath(demo_info['image_dpath']),
        '--output_dir', os.fspath(demo_info['output_dpath']),
        '--name', 'demoeval',
        '--cache_dir', 'None',
        '--loglevel', 'INFO',
        '--serial',
        '--no-db',
        '--no-viz',
    ]
    command = ' '.join(test_args)
    print(command)


if __name__ == "__main__":
    """
    CommandLine:
        python -m kwgis.demo.metrics_demo.generate_demodata --noise=0.0
    """
    main()



import os
import ubelt as ub
from kwgis.demo import sentinel2_demodata
import json


def demo():
    """
    Returns:
        str: path to a demo stac catalog
    """
    data = sentinel2_demodata.grab_sentinel2_product(0)
    gpath = os.path.join(str(data.path), str(data.images[0]))
    dpath = ub.Path.appdir('kwgis/demo/demo_stac').ensuredir()
    cat_path = os.path.join(dpath, 'demo_catalog.json')
    expected_cat_sha512 = '78e9f307e4e365c826a55f6beaa9d6e3b6b6ba4e3cc7d08'
    expected_item1_sha512 = '78224b4b2e56cbc58ca8d76b632b1c2c323eeae653b5323'
    expected_item2_sha512 = '348daaffb70c17da43f08f5b1506f929467f6986e7e8170'
    catalog = {
        "id": "demo catalog",
        "stac_version": "1.0.0-beta.2",
        "description": "demo catalog",
        "links": [
            {
                "rel": "root",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "self",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "item",
                "href": os.path.join(dpath, "testitem1.json"),
                "type": "application/json"
            },
            {
                "rel": "item",
                "href": os.path.join(dpath, "testitem2.json"),
                "type": "application/json"
            }
        ]
    }
    if (not os.path.exists(cat_path) or
            not ub.hash_file(cat_path).startswith(expected_cat_sha512)):
        with open(cat_path, 'w') as f:
            json.dump(catalog, f)
    path1 = os.path.join(dpath, "testitem1.json")
    item1 = {
        "type": "Feature",
        "stac_version": "1.0.0-beta.2",
        "id": "testitem1",
        "properties": {
            "datetime": "2018-11-06T02:17:10.060000Z",
        },
        "geometry": {},
        "links": [
            {
                "rel": "root",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "parent",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "self",
                "href": path1,
                "type": "application/json"
            }
        ],
        "assets": {
            "data": {
                "href": gpath,
                "type": "application/vnd.nitf"
            }
        },
        "bbox": [],
        "stac_extensions": []
    }
    if (not os.path.exists(path1) or
            not ub.hash_file(path1).startswith(expected_item1_sha512)):
        with open(os.path.join(dpath, "testitem1.json"), 'w') as f:
            json.dump(item1, f)
    path2 = os.path.join(dpath, "testitem2.json")
    item2 = {
        "type": "Feature",
        "stac_version": "1.0.0-beta.2",
        "id": "testitem2",
        "properties": {
            "datetime": "2018-11-06T02:17:10.060000Z",
        },
        "geometry": {},
        "links": [
            {
                "rel": "root",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "parent",
                "href": cat_path,
                "type": "application/json"
            },
            {
                "rel": "self",
                "href": path2,
                "type": "application/json"
            }
        ],
        "assets": {
            "testimg1": {
                "href": gpath,
                "type": "application/vnd.nitf",
                "roles": ["data"]
            },
            "testimg2": {
                "href": gpath,
                "type": "application/vnd.nitf",
                "roles": ["data"]
            }
        },
        "bbox": [],
        "stac_extensions": []
    }
    if (not os.path.exists(path2) or
            not ub.hash_file(path2).startswith(expected_item2_sha512)):
        with open(os.path.join(dpath, "testitem2.json"), 'w') as f:
            json.dump(item2, f)
    return cat_path


if __name__ == '__main__':
    print(demo())



import ubelt as ub
import os
import fasteners
import threading
from os.path import exists
from os.path import join


PER_PROCESS_THREAD_LOCKS = {}


def grabdata_girder(api_url, resource_id, name=None, dpath=None, hash_prefix=None,
                    appname='kwgis/girder', api_key=None,
                    verbose=1):
    """
    Downloads and caches a file or folder from girder.

    Args:
        api_url (str): the URL to the girder server

        resource_id (str): the id of the resource (e.g. file or folder).

        name (str, default=None): the desired name of the downloaded resource
            on disk if unspecified, the name of the resource on the girder
            server will be used.

        dapth (str, default=None): the path to download the resource to

        hash_prefix (str, default=None):
            If provided, and the item is a file, check that the prefix of the
            hash provided by girder matches this prefix. If the item is a
            folder, then this is not used.

        appname (str, default='kwgis'):
            if ``dpath`` is not given, then the ubelt app_cache_dir for this
            application name is used.

        api_key (str):
            Your API key to autheticate with to access private data.
            If unspecified the ``GIRDER_API_KEY`` environ is used.

        verbose (int, default=1):
            verbosity flag, if non-zero prints download progress.

    Requirements:
        pip install girder-client

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> from kwgis.utils.util_girder import *  # NOQA
        >>> from os.path import basename
        >>> api_url = 'https://data.kitware.com/api/v1'
        >>> resource_id = '59eb64678d777f31ac6477eb'
        >>> fpath = grabdata_girder(api_url, resource_id,
        >>>                         hash_prefix='79974f400aa50')
        >>> assert basename(fpath) == 'us.json'
        >>> resource_id = '59eb64298d777f31ac6477e8'
        >>> fpath = grabdata_girder(api_url, resource_id)
        >>> assert basename(fpath) == 'boundries'

    Ignore:
        import xdev
        globals().update(xdev.get_func_kwargs(grabdata_girder))
    """
    # Use the CLI version to get a progress bar
    import girder_client
    if verbose:
        from girder_client.cli import GirderCli
        client = GirderCli(username=None, password=None, apiUrl=api_url)
    else:
        client = girder_client.GirderClient(apiUrl=api_url)

    auth_info = {'api_key': api_key}
    if auth_info.get('api_key', None) is None:
        auth_info['api_key'] = os.environ.get('GIRDER_API_KEY', None)
    if auth_info.get('api_key', None) is not None:
        client.authenticate(apiKey=auth_info['api_key'])

    name = None
    dpath = None
    if dpath is None:
        dpath = ub.Path.appdir(appname).ensuredir()

    get_info_methods = {
        'file': client.getFile,
        'folder': client.getFolder,
        'item': client.getItem,
        'collection': client.getCollection,
    }

    # Determine what type of resource the requested id is for.
    resource_type = None
    for resource_type, get_info in get_info_methods.items():
        try:
            resource_info = get_info(resource_id)
        except girder_client.HttpError as ex:
            if ex.response.status_code == 400:
                pass
            elif ex.response.status_code == 401:
                print('Unauthorized, probably need to set credentials')
                raise
            else:
                print('Unhandled error ex = {}'.format(ex))
            resource_info = None
        except Exception:
            resource_info = None
        else:
            break

    if resource_info is None:
        raise Exception('Unable to determine type of resource')

    if resource_type == 'file':
        file_info = resource_info
        if name is None:
            name = file_info['name']
        dl_path = join(dpath, name)
        dl_method = client.downloadFile
        hash_value = file_info['sha512']
        if hash_prefix:
            if not hash_value.startswith(hash_prefix):
                raise ValueError('Incorrect got={}, want={}'.format(
                    hash_prefix, hash_value))

        depends = file_info['sha512']
    elif resource_type == 'item':
        raise NotImplementedError(
            'The current implementation of downloading items is '
            'not consistent. Download the file instead')
        item_info = resource_info
        dl_method = client.downloadItem
        if name is not None:
            raise ValueError('cannot specify a name for an item')
        else:
            name = item_info['name']
        dl_path = join(dpath, name)
        depends = (
            item_info['updated'],
            item_info['size'],
        )
        if hash_prefix:
            raise ValueError('Cannot specify a hash_prefix for an item')
    elif resource_type == 'folder':
        folder_info = resource_info
        dl_method = client.downloadFolderRecursive
        if name is None:
            name = folder_info['name']
        dl_path = ub.ensuredir((dpath, name))
        depends = (
            folder_info['updated'],
            folder_info['size'],
        )
        if hash_prefix:
            raise ValueError('Cannot specify a hash_prefix for a folder')
    else:
        raise KeyError(resource_type)

    cache_name = resource_type + '_' + name + '.tmp.hash'
    stamp = ub.CacheStamp(cache_name, dpath=dpath, depends=depends)
    if stamp.expired() or not exists(dl_path):
        lock_fpath = join(dpath, cache_name + '.lock')

        # Ensure that multiple processes / threads dont download the same data
        from multiprocessing import current_process
        curr_proc = current_process().name

        # Use a different thread-lock on a per-process / per-stamp basis
        lock_key = (curr_proc, lock_fpath)
        if lock_key not in PER_PROCESS_THREAD_LOCKS:
            PER_PROCESS_THREAD_LOCKS[lock_key] = threading.Lock()
        thread_lock = PER_PROCESS_THREAD_LOCKS[lock_key]

        process_lock = fasteners.InterProcessLock(lock_fpath)  #
        with thread_lock:
            with process_lock:
                if stamp.expired() or not exists(dl_path):
                    dl_method(resource_id, dl_path)
                    stamp.renew()
    return dl_path



__devnotes__ = """

# Command to autogenerate lazy imports for this file
mkinit -m kwgis.gis --lazy --noattr
mkinit -m kwgis.gis --lazy --noattr -w
"""


def lazy_import(module_name, submodules, submod_attrs):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    if os.environ.get('EAGER_IMPORT', ''):
        for name in name_to_submod.values():
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'digital_globe',
        'elevation',
        'geotiff',
        'spatial_reference',
    },
    submod_attrs={},
)


def __dir__():
    return __all__


__all__ = ['digital_globe', 'elevation', 'geotiff', 'spatial_reference']



def dummy_rpc_geotiff_fpath():
    """
    Create a blank tif with RPC information for testing
    """
    import rasterio
    import ubelt as ub
    from os.path import join
    from kwgis.gis import spatial_reference as watch_crs
    dpath = ub.Path.appdir('kwgis/demodata').ensuredir()
    gpath = join(dpath, 'test_rpc.tif')
    rpcs = watch_crs.RPCTransform.demo()

    with rasterio.open(gpath, 'w', driver='GTiff', dtype='uint8', count=1,
                       width=2000, height=2000, rpcs=rpcs.rpcs) as dst:
        dst,  # do nothing

    return gpath



import ubelt as ub
from os.path import join


def grab_landsat_product(product_id=None, demo_index=0):
    """
    Download and cache all items for a landsat product.

    Args:
        product_id (str, default=None):
            The product id to download (currently NotImplemented).
            If unspecified, an arbitrary scene is returned.

        demo_index (int):
            hack, can be 0, 1, or 2. Regions 1 and 2 should overlap.

    Returns:
        Dict[str, object]:
            groupings of files associated with this landsat product

    Example:
        >>> # xdoctest: +REQUIRES(--network)
        >>> from kwgis.demo.landsat_demodata import *  # NOQA
        >>> product = grab_landsat_product()
        >>> # xdoctest: +IGNORE_WANT
        >>> print('product = {}'.format(ub.urepr(product, nl=2)))
        product = {
            'bands': [
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B1.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B2.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B3.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B4.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B5.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B6.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B7.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B8.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B9.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B10.TIF',
                '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_B11.TIF',
            ],
            'meta': {
                'bqa': '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_BQA.TIF',
                'mtl': '.../LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1/LC08_L1TP_037029_20130602_20170310_01_T1_MTL.txt',
            },
        }

    References:
        .. [1] https://github.com/dgketchum/Landsat578#-1
        .. [2] https://tilemill-project.github.io/tilemill/docs/guides/landsat-8-imagery/
        .. [3] https://earth.esa.int/documents/700255/1834061/Landsat+ETM%2B%20Data+Format+Control+Book/4bfb7121-e97d-46ca-8d3f-02f1c6bf309c;jsessionid=F97499AA37A1E9AF0EED42900CF66760?version=1.1

    SeeAlso:
        kwgis.gis.geotiff.parse_landsat_product_id

    Ignore:
        # Use console page for the first data
        https://console.cloud.google.com/storage/browser/gcp-public-data-landsat/LC08/01/037/029/LC08_L1TP_037029_20130602_20170310_01_T1

    TODO:
        - [ ] parametarize scene name / identifier
        - [ ] bundle bands in a single file (gdal VRT?)
        - [X] separate data and metadata files in return structure?
    """
    if product_id is not None:
        raise NotImplementedError('Must use the default scene')

    if demo_index == 0:
        scene_name = 'LC08_L1TP_037029_20130602_20170310_01_T1'
        scene_path = join('LC08', '01', '037', '029', scene_name)
    elif demo_index == 1:
        scene_name = 'LC08_L1TP_187022_20191020_20191030_01_T1'
        scene_path = join('LC08', '01', '187', '022', scene_name)
    elif demo_index == 2:
        scene_name = 'LC08_L1TP_187021_20191020_20191030_01_T1'
        scene_path = join('LC08', '01', '187', '021', scene_name)
    else:
        raise KeyError(demo_index)

    if False:
        # overkill?
        from kwgis.gis.geotiff import parse_landsat_product_id
        ls_meta = parse_landsat_product_id(scene_name)
        sat_code = ls_meta['sat_code']
    else:
        # Equivalent to above, but more fragile and less dependencies
        sat_code = scene_name[2:4]

        try:
            int(sat_code.lstrip('0'))
        except Exception:
            raise AssertionError('scene name does have landsat spec')

    uri_prefix = 'http://storage.googleapis.com/gcp-public-data-landsat'

    # Each satellite has a different set of files that it will produce
    sat_code_to_suffixes = {
        '01': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF', 'B7.TIF'],
            'meta': {'mtl': 'MTL.txt'}
        },

        '02': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF', 'B7.TIF'],
            'meta': {'mtl': 'MTL.txt'}
        },

        '03': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF', 'B7.TIF'],
            'meta': {'mtl': 'MTL.txt', 'bqa': 'BQA.TIF'}
        },

        '04': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF', 'B7.TIF'],
            'meta': {'mtl': 'MTL.txt', 'bqa': 'BQA.TIF'}
        },

        '05': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF', 'B7.TIF'],
            'meta': {'mtl': 'MTL.txt', 'bqa': 'BQA.TIF'}
        },

        '07': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6_VCID_1.TIF', 'B6_VCID_2.TIF'],
            'meta': {'mtl': 'MTL.txt', 'bqa': 'BQA.TIF'}
        },

        '08': {
            'bands': ['B1.TIF', 'B2.TIF', 'B3.TIF', 'B4.TIF', 'B5.TIF', 'B6.TIF',
                      'B7.TIF', 'B8.TIF', 'B9.TIF', 'B10.TIF', 'B11.TIF'],
            'meta': {'mtl': 'MTL.txt', 'bqa': 'BQA.TIF'},
        }
    }

    # B6_VCID_1 = Band 6 Visual Channel Identifier (VCID) 1
    # B6_VCID_2 = Band 6 VCID
    # MTL = metadata
    # BQA: quality assurance pseudo-band

    item_suffixes = sat_code_to_suffixes[sat_code]

    # By default cache to the $XDG_CACHE_HOME/kwgis
    dset_dpath = ub.Path.appdir('kwgis/demo/landsat').ensuredir()

    # Cache the scene using the same path used by google cloud storage
    scene_dpath = dset_dpath / scene_path
    scene_dpath.ensuredir()

    product = {
        'bands': [],
        'meta': {},
        'scene_name': scene_name,
    }
    # Download band product-items
    for suffix in item_suffixes['bands']:
        fname = '{}_{}'.format(scene_name, suffix)
        uri_suffix = join(scene_path, fname)
        item_uri = join(uri_prefix, uri_suffix)
        fpath = ub.grabdata(item_uri, dpath=scene_dpath)

        stamp = ub.CacheStamp(fname + '.nodata.stamp', depends=[],
                              dpath=scene_dpath)
        if stamp.expired():
            import rasterio
            # TODO: cache this step
            with rasterio.open(fpath, 'r+') as img:
                if img.nodata is None:
                    img.nodata = 0
            stamp.renew()

        product['bands'].append(fpath)

    # Download meta product-items
    for key, suffix in item_suffixes['meta'].items():
        fname = '{}_{}'.format(scene_name, suffix)
        uri_suffix = join(scene_path, fname)
        item_uri = join(uri_prefix, uri_suffix)
        fpath = ub.grabdata(item_uri, dpath=scene_dpath)
        product['meta'][key] = fpath

    return product
